"""Module defining  ``Eigensolver`` classes."""

from math import pi, sqrt, sin, cos, atan2

import numpy as np
from numpy import dot
from ase.units import Hartree

from gpaw.utilities.blas import axpy
from gpaw.utilities import unpack
from gpaw.eigensolvers.eigensolver import Eigensolver


class CG(Eigensolver):
    """Conjugate gardient eigensolver

    It is expected that the trial wave functions are orthonormal
    and the integrals of projector functions and wave functions
    are already calculated.

    Solution steps are:

    * Subspace diagonalization
    * Calculate all residuals
    * Conjugate gradient steps
    """

    def __init__(self, niter=4, rtol=0.30, tw_coeff=False):
        """Construct conjugate gradient eigen solver.

        parameters:

        niter: int
            Maximum number of conjugate gradient iterations per band
        rtol: float
            If change in residual is less than rtol, iteration for band is
            not continued

        """
        Eigensolver.__init__(self)
        self.niter = niter
        self.rtol = rtol
        self.orthonormalization_required = False
        self.tw_coeff = tw_coeff

        self.tolerance = None

    def __repr__(self):
        return 'CG(niter=%d, rtol=%5.1e)' % (self.niter, self.rtol)

    def todict(self):
        return {'name': 'cg', 'niter': self.niter}

    def initialize(self, wfs):
        if wfs.bd.comm.size > 1:
            raise ValueError('CG eigensolver does not support band '
                             'parallelization.  This calculation parallelizes '
                             'over %d band groups.' % wfs.bd.comm.size)
        if wfs.mode == 'pw' and wfs.gd.comm.size > 1:
            raise ValueError('CG eigensolver does not support domain '
                             'parallelization in PW-mode.')
        Eigensolver.initialize(self, wfs)

    def iterate_one_k_point(self, ham, wfs, kpt, weights):
        """Do conjugate gradient iterations for the k-point"""
        self.timer.start('CG')

        niter = self.niter

        phi_G, phi_old_G, Htphi_G = wfs.empty(3, q=kpt.q)

        comm = wfs.gd.comm
        if self.tw_coeff:
            # Wait!  What business does the eigensolver have changing
            # the properties of the Hamiltonian?  We are not updating
            # the Hamiltonian here.  Moreover, what is supposed to
            # happen if this function is called multiple times per
            # iteration?  Then we keep dividing the potential by the
            # same number.  What on earth is the meaning of this?
            #
            # Also the parameter tw_coeff is undocumented.  What is it?
            ham.vt_sG /= self.tw_coeff
            # Assuming the ordering in dH_asp and wfs is the same
            for a in ham.dH_asp.keys():
                ham.dH_asp[a] /= self.tw_coeff

        psit = kpt.psit
        R = psit.new(buf=wfs.work_array)
        P = kpt.projections
        P2 = P.new()

        self.subspace_diagonalize(ham, wfs, kpt)

        Htpsit = psit.new(buf=self.Htpsit_nG)

        R.array[:] = Htpsit.array
        self.calculate_residuals(kpt, wfs, ham, psit,
                                 P, kpt.eps_n, R, P2)

        total_error = 0.0
        for n in range(self.nbands):
            N = self.nbands
            R_G = R.array[n]
            Htpsit_G = Htpsit.array[n]
            psit_G = psit.array[n]
            gamma_old = 1.0
            phi_old_G[:] = 0.0
            error = np.real(wfs.integrate(R_G, R_G))
            for nit in range(niter):
                if (error * Hartree**2 < self.tolerance / self.nbands):
                    break

                ekin = self.preconditioner.calculate_kinetic_energy(psit_G,
                                                                    kpt)

                pR_G = self.preconditioner(R_G, kpt, ekin)

                # New search direction
                gamma = comm.sum(np.vdot(pR_G, R_G).real)
                phi_G[:] = -pR_G - gamma / gamma_old * phi_old_G
                gamma_old = gamma
                phi_old_G[:] = phi_G[:]

                # Calculate projections
                P2_ai = wfs.pt.dict()
                wfs.pt.integrate(phi_G, P2_ai, kpt.q)

                # Orthonormalize phi_G to all bands
                self.timer.start('CG: orthonormalize')
                self.timer.start('CG: overlap')
                overlap_n = wfs.integrate(psit.array[:N], phi_G,
                                          global_integral=False)
                self.timer.stop('CG: overlap')
                self.timer.start('CG: overlap2')
                for a, P2_i in P2_ai.items():
                    P_ni = kpt.P_ani[a]
                    dO_ii = wfs.setups[a].dO_ii
                    overlap_n += np.dot(P_ni[:N].conjugate(),
                                        np.dot(dO_ii, P2_i))
                self.timer.stop('CG: overlap2')
                comm.sum(overlap_n)

                phi_G -= psit.array[:N].T.dot(overlap_n).T

                for a, P2_i in P2_ai.items():
                    P_ni = kpt.P_ani[a]
                    P2_i -= np.dot(overlap_n, P_ni[:N])

                norm = wfs.integrate(phi_G, phi_G, global_integral=False)
                for a, P2_i in P2_ai.items():
                    dO_ii = wfs.setups[a].dO_ii
                    norm += np.vdot(P2_i, np.dot(dO_ii, P2_i))
                norm = comm.sum(float(np.real(norm)))
                phi_G /= sqrt(norm)
                for P2_i in P2_ai.values():
                    P2_i /= sqrt(norm)
                self.timer.stop('CG: orthonormalize')

                # find optimum linear combination of psit_G and phi_G
                an = kpt.eps_n[n]
                wfs.apply_pseudo_hamiltonian(kpt, ham,
                                             phi_G.reshape((1,) + phi_G.shape),
                                             Htphi_G.reshape((1,) +
                                                             Htphi_G.shape))
                b = wfs.integrate(phi_G, Htpsit_G, global_integral=False)
                c = wfs.integrate(phi_G, Htphi_G, global_integral=False)
                for a, P2_i in P2_ai.items():
                    P_i = kpt.P_ani[a][n]
                    dH_ii = unpack(ham.dH_asp[a][kpt.s])
                    b += dot(P2_i, dot(dH_ii, P_i.conj()))
                    c += dot(P2_i, dot(dH_ii, P2_i.conj()))
                b = comm.sum(float(np.real(b)))
                c = comm.sum(float(np.real(c)))

                theta = 0.5 * atan2(2 * b, an - c)
                enew = (an * cos(theta)**2 +
                        c * sin(theta)**2 +
                        b * sin(2.0 * theta))
                # theta can correspond either minimum or maximum
                if (enew - kpt.eps_n[n]) > 0.0:  # we were at maximum
                    theta += pi / 2.0
                    enew = (an * cos(theta)**2 +
                            c * sin(theta)**2 +
                            b * sin(2.0 * theta))

                kpt.eps_n[n] = enew
                psit_G *= cos(theta)
                # kpt.psit_nG[n] += sin(theta) * phi_G
                axpy(sin(theta), phi_G, psit_G)
                for a, P2_i in P2_ai.items():
                    P_i = kpt.P_ani[a][n]
                    P_i *= cos(theta)
                    P_i += sin(theta) * P2_i

                if nit < niter - 1:
                    Htpsit_G *= cos(theta)
                    # Htpsit_G += sin(theta) * Htphi_G
                    axpy(sin(theta), Htphi_G, Htpsit_G)
                    # adjust residuals
                    R_G[:] = Htpsit_G - kpt.eps_n[n] * psit_G

                    coef_ai = wfs.pt.dict()
                    for a, coef_i in coef_ai.items():
                        P_i = kpt.P_ani[a][n]
                        dO_ii = wfs.setups[a].dO_ii
                        dH_ii = unpack(ham.dH_asp[a][kpt.s])
                        coef_i[:] = (dot(P_i, dH_ii) -
                                     dot(P_i * kpt.eps_n[n], dO_ii))
                    wfs.pt.add(R_G, coef_ai, kpt.q)
                    error_new = np.real(wfs.integrate(R_G, R_G))
                    if error_new / error < self.rtol:
                        # print >> self.f, "cg:iters", n, nit+1
                        break
                    if (self.nbands_converge == 'occupied' and
                        kpt.f_n is not None and kpt.f_n[n] == 0.0):
                        # print >> self.f, "cg:iters", n, nit+1
                        break
                    error = error_new

            total_error += weights[n] * error
            # if nit == 3:
            #   print >> self.f, "cg:iters", n, nit+1
        if self.tw_coeff:  # undo the scaling for calculating energies
            for i in range(len(kpt.eps_n)):
                kpt.eps_n[i] *= self.tw_coeff
            ham.vt_sG *= self.tw_coeff
            # Assuming the ordering in dH_asp and wfs is the same
            for a in ham.dH_asp.keys():
                ham.dH_asp[a] *= self.tw_coeff

        self.timer.stop('CG')
        return total_error



def quick(project='bulk', system=None):
    """Create Python script to get going quickly.
    
    project: str
        Must be 'bulk' or 'molecule'.
    system: str
        A string representing the system ('H2', 'Si').
    """
    if project == 'bulk':
        template = """\
from ase.build import bulk
from gpaw import GPAW

atoms = bulk('{0}')
atoms.calc = GPAW(kpts={{'size': (4, 4, 4)}},
                  txt='{0}.txt')
e = atoms.get_potential_energy()
f = atoms.get_forces()"""
    else:
        template = """\
from ase.build import molecule
from ase.optimize import QuasiNewton
from gpaw import GPAW

atoms = molecule('{0}')
atoms.center(vacuum=3.0)
atoms.calc = GPAW(txt='{0}.txt')
e = atoms.get_potential_energy()
opt = QuasiNewton(atoms, trajectory='{0}.traj')
opt.run(fmax=0.05)"""
    print(template.format(system))



import numpy as np
import pytest
from ase import Atoms

from gpaw import GPAW
from gpaw.mom import prepare_mom_calculation


@pytest.mark.mom
def test_mom_lcao_forces(in_tmp_dir):
    f_sn = [[1., 1., 1., 1., 0., 1., 0.],
            [1., 1., 1., 1., 1., 0., 0.]]
    L = 4.0
    d = 1.13
    delta = 0.01

    atoms = Atoms('CO',
                  [[0.5 * L, 0.5 * L, 0.5 * L - 0.5 * d],
                   [0.5 * L, 0.5 * L, 0.5 * L + 0.5 * d]])
    atoms.set_cell([L, L, L])

    calc = GPAW(mode='lcao',
                basis='dzp',
                nbands=7,
                h=0.24,
                xc='PBE',
                spinpol=True,
                symmetry='off',
                convergence={'energy': 100,
                             'density': 1e-4})

    atoms.calc = calc
    prepare_mom_calculation(calc, atoms, f_sn)
    F = atoms.get_forces()

    # Test overlaps
    calc.wfs.occupations.initialize_reference_orbitals()
    for kpt in calc.wfs.kpt_u:
        f_n = calc.get_occupation_numbers(spin=kpt.s)
        unoccupied = [True for i in range(len(f_n))]
        P = calc.wfs.occupations.calculate_weights(kpt, 1.0, unoccupied)
        assert (np.allclose(P, f_n))

    E = []
    p = atoms.positions.copy()
    for i in [-1, 1]:
        pnew = p.copy()
        pnew[0, 2] -= delta / 2. * i
        pnew[1, 2] += delta / 2. * i
        atoms.set_positions(pnew)

        E.append(atoms.get_potential_energy())

    f = np.sqrt(((F[1, :] - F[0, :])**2).sum()) * 0.5
    fnum = (E[0] - E[1]) / (2. * delta)  # central difference

    print(fnum)
    assert fnum == pytest.approx(11.52, abs=0.016)
    assert f == pytest.approx(fnum, abs=0.1)



from pathlib import Path
from typing import Union

import numpy as np
import ase.io.ulm as ulm
from ase.units import Ha

from gpaw.mpi import world, serial_comm


def ibz2bz(input_gpw: Union[Path, str],
           output_gpw: Union[Path, str] = None) -> None:
    """Unfold IBZ to full BZ and write new gpw-file.

    Example::

        ibz2bz('abc.gpw')

    This will create a 'abc-bz.gpw' file.  Works also from the command line::

        $ python3 -m gpaw.utilities.ibz2bz abc.gpw

    """
    from gpaw import GPAW

    if world.rank > 0:
        return

    input_gpw = str(input_gpw)

    calc = GPAW(input_gpw, txt=None, communicator=serial_comm)
    spos_ac = calc.atoms.get_scaled_positions()
    kd = calc.wfs.kd

    nK = kd.N_c.prod()
    nbands = calc.wfs.bd.nbands
    ns = calc.wfs.nspins

    I_a = None

    for K, k in enumerate(kd.bz2ibz_k):
        a_a, U_aii, time_rev = construct_symmetry_operators(
            calc.wfs.kd, calc.wfs.setups, spos_ac, K)

        if I_a is None:
            # First time:
            I1 = 0
            I_a = [0]
            for U_ii in U_aii:
                I2 = I1 + len(U_ii)
                I_a.append(I2)
                I1 = I2

            # Allocate arrays:
            P_sKnI = np.empty((ns, nK, nbands, I2), calc.wfs.dtype)
            e_sKn = np.empty((ns, nK, nbands))
            f_sKn = np.empty((ns, nK, nbands))

        weight = kd.weight_k[k] * (2 // kd.nspins)
        for s in range(ns):
            e_sKn[s, K] = calc.get_eigenvalues(k, s)
            f_sKn[s, K] = calc.get_occupation_numbers(k, s) / weight
            P_nI = calc.wfs.collect_projections(k, s)
            P2_nI = P_sKnI[s, K]
            a = 0
            for b, U_ii in zip(a_a, U_aii):
                P_ni = np.dot(P_nI[:, I_a[b]:I_a[b + 1]], U_ii)
                if time_rev:
                    P_ni = P_ni.conj()
                P2_nI[:, I_a[a]:I_a[a + 1]] = P_ni
                a += 1

    # Write new gpw-file:
    parameters = calc.todict()
    parameters['symmetry'] = 'off'

    if output_gpw is None:
        assert input_gpw.endswith('.gpw')
        output_gpw = input_gpw[:-4] + '-bz.gpw'
    out = ulm.open(output_gpw, 'w')

    ulm.copy(input_gpw, out, exclude={'.wave_functions', '.parameters'})

    out.child('parameters').write(**parameters)

    wfs = out.child('wave_functions')
    wfs.write(eigenvalues=e_sKn,
              fermi_levels=calc.wfs.fermi_levels * Ha,
              occupations=f_sKn,
              projections=P_sKnI)

    wfs.child('kpts').write(bz2ibz=np.arange(nK),
                            bzkpts=kd.bzk_kc,
                            ibzkpts=kd.bzk_kc,
                            weights=np.ones(nK) / nK)
    out.close()


def construct_symmetry_operators(kd, setups, spos_ac, K):
    """Construct symmetry operators for PAW projections.

    We want to transform a k-point in the irreducible part of the BZ to
    the corresponding k-point with index K.

    Returns a_a, U_aii, and time_reversal, where:

    * a_a is a list of symmetry related atom indices
    * U_aii is a list of rotation matrices for the PAW projections
    * time_reversal is a flag - if True, projections should be complex
      conjugated.
    """

    s = kd.sym_k[K]
    U_cc = kd.symmetry.op_scc[s]
    time_reversal = kd.time_reversal_k[K]
    ik = kd.bz2ibz_k[K]
    ik_c = kd.ibzk_kc[ik]

    a_a = []
    U_aii = []
    for a, id in enumerate(setups.id_a):
        b = kd.symmetry.a_sa[s, a]
        S_c = np.dot(spos_ac[a], U_cc) - spos_ac[b]
        x = np.exp(2j * np.pi * np.dot(ik_c, S_c))
        U_ii = setups[a].R_sii[s].T * x
        a_a.append(b)
        U_aii.append(U_ii)

    return a_a, U_aii, time_reversal


if __name__ == '__main__':
    import sys
    ibz2bz(sys.argv[1])



from types import SimpleNamespace

import numpy as np

import gpaw.gpu.cpupy.cublas as cublas
import gpaw.gpu.cpupy.fft as fft
import gpaw.gpu.cpupy.linalg as linalg

__all__ = ['linalg', 'cublas', 'fft']


def empty(*args, **kwargs):
    return ndarray(np.empty(*args, **kwargs))


def empty_like(a):
    return ndarray(np.empty_like(a._data))


def zeros(*args, **kwargs):
    return ndarray(np.zeros(*args, **kwargs))


def ones(*args, **kwargs):
    return ndarray(np.ones(*args, **kwargs))


def asnumpy(a, out=None):
    if out is None:
        return a._data.copy()
    out[:] = a._data
    return out


def asarray(a):
    if isinstance(a, ndarray):
        return a
    return ndarray(np.array(a))


def array(a):
    return ndarray(np.array(a))


def multiply(a, b, c):
    np.multiply(a._data, b._data, c._data)


def negative(a, b):
    np.negative(a._data, b._data)


def einsum(indices, *args):
    return ndarray(
        np.einsum(
            indices,
            *(arg._data for arg in args)))


def diag(a):
    return ndarray(np.diag(a._data))


def abs(a):
    return ndarray(np.abs(a._data))


def exp(a):
    return ndarray(np.exp(a._data))


def eye(n):
    return ndarray(np.eye(n))


def triu_indices(n, k=0, m=None):
    i, j = np.triu_indices(n, k, m)
    return ndarray(i), ndarray(j)


def moveaxis(a, source, destination):
    return ndarray(np.moveaxis(a._data, source, destination))


def vdot(a, b):
    return np.vdot(a._data, b._data)


def fuse():
    return lambda func: func


class ndarray:
    def __init__(self, data):
        if isinstance(data, (float, complex, int, np.int32, np.int64,
                             np.bool_)):
            data = np.asarray(data)
        assert isinstance(data, np.ndarray), type(data)
        self._data = data
        self.dtype = data.dtype
        self.size = data.size
        self.flags = data.flags
        self.ndim = data.ndim
        self.nbytes = data.nbytes
        self.data = SimpleNamespace(ptr=data.ctypes.data)

    @property
    def shape(self):
        return self._data.shape

    @property
    def T(self):
        return ndarray(self._data.T)

    @property
    def real(self):
        return ndarray(self._data.real)

    @property
    def imag(self):
        return ndarray(self._data.imag)

    def set(self, a):
        if self.ndim == 0:
            self._data.fill(a)
        else:
            self._data[:] = a

    def get(self):
        return self._data.copy()

    def copy(self):
        return ndarray(self._data.copy())

    def all(self):
        return ndarray(self._data.all())

    def sum(self, **kwargs):
        return ndarray(self._data.sum(**kwargs))

    def __len__(self):
        return len(self._data)

    def __bool__(self):
        return bool(self._data)

    def __float__(self):
        return self._data.__float__()

    def __iter__(self):
        for data in self._data:
            if data.ndim == 0:
                yield ndarray(data.item())
            else:
                yield ndarray(data)

    def __setitem__(self, index, value):
        if isinstance(index, tuple):
            def convert(a):
                return a._data if isinstance(a, ndarray) else a
            index = tuple([convert(a) for a in index])
        if isinstance(index, ndarray):
            index = index._data
        if isinstance(value, ndarray):
            self._data[index] = value._data
        else:
            assert isinstance(value, (float, int, complex))
            self._data[index] = value

    def __getitem__(self, index):
        if isinstance(index, tuple):
            def convert(a):
                return a._data if isinstance(a, ndarray) else a
            index = tuple([convert(a) for a in index])
        if isinstance(index, ndarray):
            index = index._data
        return ndarray(self._data[index])

    def __eq__(self, other):
        if isinstance(other, (float, int)):
            return self._data == other
        return ndarray(self._data == other._data)

    def __mul__(self, f):
        if isinstance(f, (float, complex)):
            return ndarray(f * self._data)
        return ndarray(f._data * self._data)

    def __rmul__(self, f):
        return ndarray(f * self._data)

    def __imul__(self, f):
        if isinstance(f, (float, complex, int)):
            self._data *= f
        else:
            self._data *= f._data
        return self

    def __truediv__(self, other):
        if isinstance(other, (float, complex, int)):
            return ndarray(self._data / other)
        return ndarray(self._data / other._data)

    def __pow__(self, i: int):
        return ndarray(self._data**i)

    def __add__(self, f):
        if isinstance(f, (float, int, complex)):
            return ndarray(f + self._data)
        return ndarray(f._data + self._data)

    def __sub__(self, f):
        if isinstance(f, float):
            return ndarray(self._data - f)
        return ndarray(self._data - f._data)

    def __rsub__(self, f):
        return ndarray(f - self._data)

    def __radd__(self, f):
        return ndarray(f + self._data)

    def __rtruediv__(self, f):
        return ndarray(f / self._data)

    def __iadd__(self, other):
        if isinstance(other, float):
            self._data += other
        else:
            self._data += other._data
        return self

    def __isub__(self, other):
        self._data -= other._data
        return self

    def __matmul__(self, other):
        return ndarray(self._data @ other._data)

    def ravel(self):
        return ndarray(self._data.ravel())

    def conj(self):
        return ndarray(self._data.conj())

    def reshape(self, shape):
        return ndarray(self._data.reshape(shape))

    def view(self, dtype):
        return ndarray(self._data.view(dtype))

    def item(self):
        return self._data.item()

    def trace(self, offset, axis1, axis2):
        return ndarray(self._data.trace(offset, axis1, axis2))



import subprocess
from pathlib import Path
from typing import Union, IO, Dict, Any, cast

from ase import Atoms
import numpy as np

from .overlaps import WannierOverlaps
from .functions import WannierFunctions
from gpaw.typing import Array3D


class Wannier90Error(Exception):
    """Wannier90 error."""


class Wannier90:
    def __init__(self,
                 prefix: str = 'wannier',
                 folder: Union[str, Path] = 'W90',
                 executable='wannier90.x'):
        self.prefix = prefix
        self.folder = Path(folder)
        self.executable = executable
        self.folder.mkdir(exist_ok=True)

    def run_wannier90(self, postprocess=False, world=None):
        args = [self.executable, self.prefix]
        if postprocess:
            args[1:1] = ['-pp']
        result = subprocess.run(args,
                                cwd=self.folder,
                                stdout=subprocess.PIPE)
        if b'Error:' in result.stdout:
            raise Wannier90Error(result.stdout.decode())

    def write_input_files(self,
                          overlaps: WannierOverlaps,
                          **kwargs) -> None:
        self.write_win(overlaps, **kwargs)
        self.write_mmn(overlaps)
        if overlaps.projections is not None:
            self.write_amn(overlaps.projections)

    def write_win(self,
                  overlaps: WannierOverlaps,

                  **kwargs) -> None:
        kwargs['num_bands'] = overlaps.nbands
        kwargs['num_wann'] = overlaps.nwannier
        kwargs['fermi_energy'] = overlaps.fermi_level
        kwargs['unit_cell_cart'] = overlaps.atoms.cell.tolist()
        kwargs['atoms_frac'] = [[symbol] + list(spos_c)
                                for symbol, spos_c
                                in zip(overlaps.atoms.symbols,
                                       overlaps.atoms.get_scaled_positions())]
        kwargs['mp_grid'] = tuple(overlaps.monkhorst_pack_size)
        kwargs['kpoints'] = overlaps.kpoints
        if overlaps.proj_indices_a:
            kwargs['guiding_centres'] = True
            centers = []
            for (x, y, z), indices in zip(overlaps.atoms.positions,
                                          overlaps.proj_indices_a):
                centers += [[f'c={x},{y},{z}: s']] * len(indices)
            kwargs['projections'] = centers

        with (self.folder / f'{self.prefix}.win').open('w') as fd:
            for key, val in kwargs.items():
                if isinstance(val, tuple):
                    print(f'{key} =', *val, file=fd)
                elif isinstance(val, (list, np.ndarray)):
                    print(f'begin {key}', file=fd)
                    for line in val:
                        print('   ', *line, file=fd)
                    print(f'end {key}', file=fd)
                else:
                    print(f'{key} = {val}', file=fd)

    def write_mmn(self,
                  overlaps: WannierOverlaps) -> None:
        size = overlaps.monkhorst_pack_size
        nbzkpts = cast(int, np.prod(size))
        nbands = overlaps.nbands

        directions = list(overlaps.directions)
        directions += [(-a, -b, -c) for (a, b, c) in directions]
        ndirections = len(directions)

        with (self.folder / f'{self.prefix}.mmn').open('w') as fd:
            print('Input generated from GPAW', file=fd)
            print(f'{nbands} {nbzkpts} {ndirections}', file=fd)

            for bz_index1 in range(nbzkpts):
                i1_c = np.unravel_index(bz_index1, size)
                for direction in directions:
                    i2_c = np.array(i1_c) + direction
                    bz_index2 = np.ravel_multi_index(i2_c,
                                                     size,
                                                     'wrap')  # type: ignore
                    d_c = (i2_c - i2_c % size) // size
                    print(bz_index1 + 1, bz_index2 + 1, *d_c, file=fd)
                    M_nn = overlaps.overlap(bz_index1, direction)
                    for M_n in M_nn.T:
                        for M in M_n:
                            print(f'{M.real} {M.imag}', file=fd)

    def write_amn(self,
                  proj_kmn: Array3D) -> None:
        nbzkpts, nproj, nbands = proj_kmn.shape

        with (self.folder / f'{self.prefix}.amn').open('w') as fd:
            print('Input generated from GPAW', file=fd)
            print(f'{nbands} {nbzkpts} {nproj}', file=fd)

            for bz_index, proj_mn in enumerate(proj_kmn):
                for m, proj_n in enumerate(proj_mn):
                    for n, P in enumerate(proj_n):
                        print(n + 1, m + 1, bz_index + 1, P.real, -P.imag,
                              file=fd)

    def read_result(self):
        with (self.folder / f'{self.prefix}.wout').open() as fd:
            w = read_wout_all(fd)
        return Wannier90Functions(w['atoms'], w['centers'])


class Wannier90Functions(WannierFunctions):
    def __init__(self,
                 atoms: Atoms,
                 centers):
        WannierFunctions.__init__(self, atoms, centers, 0.0, [])


def read_wout_all(fileobj: IO[str]) -> Dict[str, Any]:
    """Read atoms, wannier function centers and spreads."""
    lines = fileobj.readlines()

    for n, line in enumerate(lines):
        if line.strip().lower().startswith('lattice vectors (ang)'):
            break
    else:
        raise ValueError('Could not fine lattice vectors')

    cell = [[float(x) for x in line.split()[-3:]]
            for line in lines[n + 1:n + 4]]

    for n, line in enumerate(lines):
        if 'cartesian coordinate (ang)' in line.lower():
            break
    else:
        raise ValueError('Could not find coordinates')

    positions = []
    symbols = []
    n += 2
    while True:
        words = lines[n].split()
        if len(words) == 1:
            break
        positions.append([float(x) for x in words[-4:-1]])
        symbols.append(words[1])
        n += 1

    atoms = Atoms(symbols, positions, cell=cell, pbc=True)

    n = len(lines) - 1
    while n > 0:
        if lines[n].strip().lower().startswith('final state'):
            break
        n -= 1
    else:
        return {'atoms': atoms,
                'centers': np.zeros((0, 3)),
                'spreads': np.zeros((0,))}

    n += 1
    centers = []
    spreads = []
    while True:
        line = lines[n].strip()
        if line.startswith('WF'):
            centers.append([float(x)
                            for x in
                            line.split('(')[1].split(')')[0].split(',')])
            spreads.append(float(line.split()[-1]))
            n += 1
        else:
            break

    return {'atoms': atoms,
            'centers': np.array(centers),
            'spreads': np.array(spreads)}



import ase.db
from ase.build import molecule
from ase.optimize.bfgs import BFGS
from ase.calculators.nwchem import NWChem
from ase.data.g2_1 import molecule_names, atom_names


c = ase.db.connect('g2-1.db')

for name in molecule_names + atom_names:
    id = c.reserve(name=name, calculator='nwchem')
    if id is None:
        continue
    atoms = molecule(name)
    atoms.calc = NWChem(command='mpiexec -np 4 nwchem PREFIX.nw > PREFIX.out',
                        geometry='noautosym nocenter noautoz',
                        task='gradient',
                        xc='PBE',
                        grid='nodisk',
                        tolerances='tight',
                        basis='def2-qzvppd',
                        basispar='spherical',
                        direct='noio',
                        label=name)
    atoms.get_forces()
    c.write(atoms, name=name, relaxed=False)
    if len(atoms) > 1:
        opt = BFGS(atoms, logfile=name + '.nwchem.log')
        opt.run(0.01)
        c.write(atoms, name=name, relaxed=True)
    del c[id]



from math import pi
import numpy as np
import pytest
from gpaw.xc.libxc import LibXC


@pytest.mark.libxc
def test_xc_pbe_pw91():
    x0 = LibXC('LDA_X')

    def f0(xc, rs, s):
        n = 3 / (4 * pi * rs**3)
        third = 1.0 / 3.0
        kF = (3 * pi**2 * n)**third
        a2 = (2 * kF * n * s)**2
        e = np.zeros(1)
        xc.calculate(e,
                     np.array([[n]]), np.zeros((1, 1)),
                     np.array([[a2]]), np.zeros((1, 1)))
        exc = n * e[0]
        x0.calculate(e, np.array([[n]]), np.zeros((1, 1)))
        ex0 = n * e[0]
        return exc / ex0

    def f1(xc, rs, s):
        n = 3 / (4 * pi * rs**3)
        # na = 2 * n
        third = 1.0 / 3.0
        kF = (3 * pi**2 * n)**third
        a2 = (2 * kF * n * s)**2
        e = np.zeros(1)
        xc.calculate(e,
                     np.array([[n], [0]]), np.zeros((2, 1)),
                     np.array([[a2], [0], [0]]), np.zeros((3, 1)))
        exc = n * e[0]
        x0.calculate(e, np.array([[n]]), np.zeros((1, 1)))
        ex0 = n * e[0]
        return exc / ex0

    pbe = LibXC('PBE')
    pw91 = LibXC('PW91')
    assert abs(f0(pbe, 2, 3) - 1.58) < 0.01
    assert abs(f1(pbe, 2, 3) - 1.88) < 0.01
    assert abs(f0(pw91, 2, 3) - 1.60) < 0.01
    assert abs(f1(pw91, 2, 3) - 1.90) < 0.01



import pytest
from ase import Atoms, Atom
from ase.vibrations.resonant_raman import ResonantRaman
from ase.vibrations.albrecht import Albrecht
from ase.parallel import world, DummyMPI

from gpaw import GPAW
from gpaw.lrtddft.kssingle import KSSingles
from gpaw.analyse.overlap import Overlap
from gpaw.test import equal


@pytest.mark.skip(reason='TODO')
def test_rraman_albrecht():
    txt = '-'
    txt = None
    xc = 'LDA'

    R = 0.7  # approx. experimental bond length
    a = 4.0
    c = 5.0
    H2 = Atoms([Atom('H', (a / 2, a / 2, (c - R) / 2)),
                Atom('H', (a / 2, a / 2, (c + R) / 2))],
               cell=(a, a, c))

    calc = GPAW(xc=xc, nbands=3, spinpol=False, eigensolver='rmm-diis',
                txt=txt)
    H2.calc = calc
    H2.get_potential_energy()

    gsname = exname = 'rraman'
    exkwargs = {'eps': 0.0, 'jend': 1}
    pz = ResonantRaman(H2, KSSingles, gsname=gsname, exname=exname,
                       exkwargs=exkwargs,
                       overlap=lambda x, y: Overlap(x).pseudo(y))
    pz.run()

    # check size
    kss = KSSingles('rraman-d0.010.eq.ex.gz')
    assert len(kss) == 1

    om = 5

    # Does Albrecht A work at all ?
    # -----------------------------

    al = Albrecht(H2, KSSingles, gsname=gsname, exname=exname,
                  verbose=True, overlap=True)
    _ = al.absolute_intensity(omega=om)[-1]
    i = al.intensity(omega=om)[-1]

    # parallel

    if world.size > 1 and world.rank == 0:
        # single core
        comm = DummyMPI()
        pzsi = Albrecht(H2, KSSingles, gsname=gsname, exname=exname,
                        comm=comm, overlap=True, verbose=True)
        isi = pzsi.intensity(omega=om)[-1]
        equal(isi, i, 1e-11)

    # Compare singles and multiples in Albrecht A
    # -------------------------------------------

    alas = Albrecht(H2, KSSingles, gsname=gsname, exname=exname,
                    approximation='Albrecht A')
    ints = alas.intensity(omega=om)[-1]
    alam = Albrecht(H2, KSSingles, gsname=gsname, exname=exname,
                    approximation='Albrecht A',
                    skip=5, combinations=2)
    # single excitation energies should agree
    equal(ints, alam.intensity(omega=om)[0], 1e-11)
    alam.summary(omega=5, method='frederiksen')



from typing import Callable

import _gpaw
import numpy as np
from gpaw.core.plane_waves import PlaneWaveExpansions
from gpaw.core.uniform_grid import UniformGridFunctions
from gpaw.gpu import cupy as cp
from gpaw.new import zip
from gpaw.new.hamiltonian import Hamiltonian


class PWHamiltonian(Hamiltonian):
    def __init__(self, grid, pw, xp):
        self.plan = grid.new(dtype=pw.dtype).fft_plans(xp=xp)
        self.pw_cache = {}

    def apply(self,
              vt_sR: UniformGridFunctions,
              psit_nG: PlaneWaveExpansions,
              out: PlaneWaveExpansions,
              spin: int) -> PlaneWaveExpansions:
        out_nG = out
        vt_R = vt_sR[spin].gather(broadcast=True)
        xp = psit_nG.xp
        grid = vt_R.desc.new(comm=None, dtype=psit_nG.desc.dtype)
        tmp_R = grid.empty(xp=xp)
        pw = psit_nG.desc
        if pw.comm.size == 1:
            pw_local = pw
        else:
            key = tuple(pw.kpt_c)
            pw_local = self.pw_cache.get(key)
            if pw_local is None:
                pw_local = pw.new(comm=None)
                self.pw_cache[key] = pw_local
        psit_G = pw_local.empty(xp=xp)
        e_kin_G = xp.asarray(psit_G.desc.ekin_G)
        domain_comm = psit_nG.desc.comm
        mynbands = psit_nG.mydims[0]
        for n1 in range(0, mynbands, domain_comm.size):
            n2 = min(n1 + domain_comm.size, mynbands)
            psit_nG[n1:n2].gather_all(psit_G)
            psit_G.ifft(out=tmp_R)
            tmp_R.data *= vt_R.data
            vtpsit_G = tmp_R.fft(pw=psit_G.desc)
            psit_G.data *= e_kin_G
            vtpsit_G.data += psit_G.data
            out_nG[n1:n2].scatter_from_all(vtpsit_G)
        return out_nG

    def create_preconditioner(self,
                              blocksize: int
                              ) -> Callable[[PlaneWaveExpansions,
                                             PlaneWaveExpansions,
                                             PlaneWaveExpansions], None]:
        return precondition


def precondition(psit_nG: PlaneWaveExpansions,
                 residual_nG: PlaneWaveExpansions,
                 out: PlaneWaveExpansions) -> None:
    """Preconditioner for KS equation.

    From:

      Teter, Payne and Allen, Phys. Rev. B 40, 12255 (1989)

    as modified by:

      Kresse and Furthmüller, Phys. Rev. B 54, 11169 (1996)
    """

    xp = psit_nG.xp
    G2_G = xp.asarray(psit_nG.desc.ekin_G * 2)
    ekin_n = psit_nG.norm2('kinetic')

    if xp is np:
        for r_G, o_G, ekin in zip(residual_nG.data,
                                  out.data,
                                  ekin_n):
            _gpaw.pw_precond(G2_G, r_G, ekin, o_G)
        return

    out.data[:] = gpu_prec(ekin_n[:, np.newaxis],
                           G2_G[np.newaxis],
                           residual_nG.data)


@cp.fuse()
def gpu_prec(ekin, G2, residual):
    x = 1 / ekin / 3 * G2
    a = 27.0 + x * (18.0 + x * (12.0 + x * 8.0))
    xx = x * x
    return -4.0 / 3 / ekin * a / (a + 16.0 * xx * xx) * residual


def spinor_precondition(psit_nsG, residual_nsG, out):
    G2_G = psit_nsG.desc.ekin_G * 2
    for r_sG, o_sG, ekin in zip(residual_nsG.data,
                                out.data,
                                psit_nsG.norm2('kinetic').sum(1)):
        for r_G, o_G in zip(r_sG, o_sG):
            _gpaw.pw_precond(G2_G, r_G, ekin, o_G)


class SpinorPWHamiltonian(Hamiltonian):
    def apply(self,
              vt_xR: UniformGridFunctions,
              psit_nsG: PlaneWaveExpansions,
              out: PlaneWaveExpansions,
              spin: int):
        out_nsG = out
        pw = psit_nsG.desc

        if pw.qspiral_v is None:
            np.multiply(pw.ekin_G, psit_nsG.data, out_nsG.data)
        else:
            for s, sign in enumerate([1, -1]):
                ekin_G = 0.5 * ((pw.G_plus_k_Gv +
                                 0.5 * sign * pw.qspiral_v)**2).sum(1)
                np.multiply(ekin_G, psit_nsG.data[:, s], out_nsG.data[:, s])

        grid = vt_xR.desc.new(dtype=complex)

        v, x, y, z = vt_xR.data
        iy = y * 1j

        f_sR = grid.empty(2)
        g_R = grid.empty()

        for p_sG, o_sG in zip(psit_nsG, out_nsG):
            p_sG.ifft(out=f_sR)
            a, b = f_sR.data
            g_R.data = a * (v + z) + b * (x - iy)
            o_sG.data[0] += g_R.fft(pw=pw).data
            g_R.data = a * (x + iy) + b * (v - z)
            o_sG.data[1] += g_R.fft(pw=pw).data

        return out_nsG

    def create_preconditioner(self, blocksize):
        return spinor_precondition



"""This module defines a linear response TDDFT-class."""
import numbers
import sys
from math import sqrt
from typing import Dict, Any
import numpy as np

from ase.units import Hartree
from ase.utils.timing import Timer

import _gpaw
import gpaw.mpi as mpi
from gpaw.xc import XC
from gpaw.lrtddft.excitation import Excitation, ExcitationList, get_filehandle
from gpaw.lrtddft.kssingle import KSSingles
from gpaw.lrtddft.omega_matrix import OmegaMatrix
from gpaw.lrtddft.apmb import ApmB
from gpaw.lrtddft.spectrum import spectrum

__all__ = ['LrTDDFT', 'photoabsorption_spectrum', 'spectrum']


class LrTDDFT(ExcitationList):
    """Linear Response TDDFT excitation list class

    Input parameters:

    calculator:
    the calculator object after a ground state calculation

    nspins:
    number of spins considered in the calculation
    Note: Valid only for unpolarised ground state calculation

    eps:
    Minimal occupation difference for a transition (default 0.001)

    istart:
    First occupied state to consider
    jend:
    Last unoccupied state to consider

    xc:
    Exchange-Correlation approximation in the Kernel
    derivative_level:
    0: use Exc, 1: use vxc, 2: use fxc  if available

    filename:
    read from a file
    """

    default_parameters: Dict[str, Any] = {
        'nspins': None,
        'restrict': {},
        'xc': None,
        'derivative_level': 1,
        'numscale': 0.00001,
        'filename': None,
        'finegrid': 2,
        'force_ApmB': False,  # for tests
        'eh_comm': None,  # parallelization over eh-pairs
        'poisson': None}  # use calculator's Poisson

    def __init__(self, calculator=None, log=None, txt='-', **kwargs):

        self.energy_to_eV_scale = Hartree
        self.timer = Timer()
        self.diagonalized = False

        self.set(**kwargs)
        self.calculator = calculator

        ExcitationList.__init__(self, log=log, txt=txt)

        if self.eh_comm is None:
            self.eh_comm = mpi.serial_comm
        elif isinstance(self.eh_comm, (mpi.world.__class__,
                                       mpi.serial_comm.__class__)):
            # Correct type already.
            pass
        else:
            # world should be a list of ranks:
            self.eh_comm = mpi.world.new_communicator(
                np.asarray(self.eh_comm))

        if calculator is not None and calculator.initialized:
            # XXXX not ready for k-points
            assert len(calculator.wfs.kd.bzk_kc) == 1
            if calculator.wfs.mode not in ['fd', 'lcao']:
                raise RuntimeError('LrTDDFT supports only fd and lcao modes')
            if calculator.wfs.mode != 'lcao':
                calculator.converge_wave_functions()
            if calculator.density.nct_G is None:
                spos_ac = calculator.initialize_positions()
                calculator.wfs.initialize(calculator.density,
                                          calculator.hamiltonian, spos_ac)

            self.forced_update()

    @property
    def calculator(self):
        return self._calc

    @calculator.setter
    def calculator(self, calc):
        self._calc = calc

        if self.xc is None and calc is not None:
            if calc.initialized:
                self.xc = calc.hamiltonian.xc
            else:
                self.xc = calc.parameters['xc']

    def set(self, **kwargs):
        """Change parameters."""
        changed = []
        for key, value in LrTDDFT.default_parameters.items():
            if hasattr(self, key):
                value = getattr(self, key)  # do not overwrite
            setattr(self, key, kwargs.pop(key, value))
            if value != getattr(self, key):
                changed.append(key)

        for key in kwargs:
            raise KeyError('Unknown key ' + key)

        return changed

    def analyse(self, what=None, out=None, min=0.1):
        """Print info about the transitions.

        Parameters:
          1. what: I list of excitation indicees, None means all
          2. out : I where to send the output, None means sys.stdout
          3. min : I minimal contribution to list (0<min<1)
        """
        if what is None:
            what = range(len(self))
        elif isinstance(what, numbers.Integral):
            what = [what]

        if out is None:
            out = sys.stdout

        for i in what:
            print(str(i) + ':', self[i].analyse(min=min), file=out)

    def calculate(self, atoms):
        self.calculator = atoms.calc
        if not hasattr(self, 'Om') or self.calculator.check_state(atoms):
            self.calculator.get_potential_energy(atoms)
            self.forced_update()
        return self

    def forced_update(self):
        """Recalc yourself."""
        if not self.force_ApmB:
            Om = OmegaMatrix
            name = 'LrTDDFT'
            if self.xc:
                if isinstance(self.xc, str):
                    xc = XC(self.xc)
                else:
                    xc = self.xc
                if hasattr(xc, 'hybrid') and xc.hybrid > 0.0:
                    Om = ApmB
                    name = 'LrTDDFThyb'
        else:
            Om = ApmB
            name = 'LrTDDFThyb'

        kss = KSSingles(restrict=self.restrict,
                        log=self.log)
        atoms = self.calculator.get_atoms()
        kss.calculate(atoms, self.nspins)

        self.Om = Om(self.calculator, kss,
                     self.xc, self.derivative_level, self.numscale,
                     finegrid=self.finegrid, eh_comm=self.eh_comm,
                     poisson=self.poisson, log=self.log)
        self.name = name

    def diagonalize(self, **kwargs):
        """Diagonalize and save new Eigenvalues and Eigenvectors"""
        self.set(**kwargs)
        self.timer.start('diagonalize')
        self.timer.start('omega')
        self.Om.diagonalize(kwargs.pop('restrict', {}))
        self.timer.stop('omega')
        self.diagonalized = True

        # remove old stuff
        self.timer.start('clean')
        while len(self):
            self.pop()
        self.timer.stop('clean')

        self.log('LrTDDFT digonalized:')
        self.timer.start('build')
        for j in range(len(self.Om.kss)):
            self.append(LrTDDFTExcitation(self.Om, j))
            self.log(' ', str(self[-1]))
        self.timer.stop('build')
        self.timer.stop('diagonalize')

    @classmethod
    def read(cls, filename=None, fh=None, restrict={}, log=None, txt=None):
        """Read myself from a file"""
        lr = cls(log=log, txt=txt)
        timer = lr.timer
        timer.start('name')
        if fh is None:
            f = get_filehandle(lr, filename)
        else:
            f = fh
        timer.stop('name')

        timer.start('header')
        # get my name
        s = f.readline().strip()
        lr.name = s.split()[1]

        lr.xc = XC(f.readline().strip().split()[0])
        values = f.readline().split()
        eps = float(values[0])
        if len(values) > 1:
            lr.derivative_level = int(values[1])
            lr.numscale = float(values[2])
            lr.finegrid = int(values[3])
        else:
            # old writing style, use old defaults
            lr.numscale = 0.001
        timer.stop('header')

        timer.start('init_kss')
        kss = KSSingles.read(fh=f, log=log)
        assert eps == kss.restrict['eps']
        lr.restrict = kss.restrict.values
        timer.stop('init_kss')
        timer.start('init_obj')
        if lr.name == 'LrTDDFT':
            lr.Om = OmegaMatrix(kss=kss, filehandle=f, log=lr.log)
        else:
            lr.Om = ApmB(kss=kss, filehandle=f, log=lr.log)
        timer.stop('init_obj')

        if not len(restrict):
            timer.start('read diagonalized')
            # check if already diagonalized
            p = f.tell()
            s = f.readline()
            if s != '# Eigenvalues\n' or len(restrict):
                # no further info or selection of
                # Kohn-Sham states changed
                # go back to previous position
                f.seek(p)
            else:
                lr.diagonalized = True
                # load the eigenvalues
                n = int(f.readline().split()[0])
                for i in range(n):
                    lr.append(LrTDDFTExcitation(string=f.readline()))
                # load the eigenvectors
                timer.start('read eigenvectors')
                f.readline()
                for i in range(n):
                    lr[i].f = np.array([float(x) for x in
                                        f.readline().split()])
                    lr[i].kss = lr.kss
                timer.stop('read eigenvectors')
            timer.stop('read diagonalized')
        else:
            timer.start('diagonalize')
            lr.diagonalize(restrict=restrict)
            timer.stop('diagonalize')

        if fh is None:
            f.close()

        return lr

    @property
    def kss(self):
        return self.Om.kss

    def singlets_triplets(self):
        """Split yourself into a singlet and triplet object"""

        slr = LrTDDFT(nspins=self.nspins, xc=self.xc,
                      restrict=self.kss.restrict.values,
                      derivative_level=self.derivative_level,
                      numscale=self.numscale)
        tlr = LrTDDFT(nspins=self.nspins, xc=self.xc,
                      restrict=self.kss.restrict.values,
                      derivative_level=self.derivative_level,
                      numscale=self.numscale)
        slr.Om, tlr.Om = self.Om.singlets_triplets()

        return slr, tlr

    def single_pole_approximation(self, i, j):
        """Return the excitation according to the
        single pole approximation. See e.g.:
        Grabo et al, Theochem 501 (2000) 353-367
        """
        for ij, kss in enumerate(self.kss):
            if kss.i == i and kss.j == j:
                return sqrt(self.Om.full[ij][ij]) * Hartree
                return self.Om.full[ij][ij] / kss.energy * Hartree

    def __str__(self):
        string = ExcitationList.__str__(self)
        string += '# derived from:\n'
        string += self.Om.kss.__str__()
        return string

    def write(self, filename=None, fh=None):
        """Write current state to a file.

        'filename' is the filename. If the filename ends in .gz,
        the file is automatically saved in compressed gzip format.

        'fh' is a filehandle. This can be used to write into already
        opened files.
        """

        if self.calculator is None:
            rank = mpi.world.rank
        else:
            rank = self.calculator.wfs.world.rank

        if rank == 0:
            if fh is None:
                f = get_filehandle(self, filename, mode='w')
            else:
                f = fh

            f.write('# ' + self.name + '\n')
            if isinstance(self.xc, str):
                xc = self.xc
            else:
                xc = self.xc.tostring()
            if xc is None:
                xc = 'RPA'
            if self.calculator is not None:
                xc += ' ' + self.calculator.get_xc_functional()
            f.write(xc + '\n')
            f.write('%g %d %g %d' % (self.kss.restrict['eps'],
                                     int(self.derivative_level),
                                     self.numscale, int(self.finegrid)) + '\n')
            self.kss.write(fh=f)
            self.Om.write(fh=f)

            if len(self):
                f.write('# Eigenvalues\n')
                f.write('{0}\n'.format(len(self)))
                for ex in self:
                    f.write(ex.outstring())
                f.write('# Eigenvectors\n')
                for ex in self:
                    for w in ex.f:
                        f.write('%g ' % w)
                    f.write('\n')

            if fh is None:
                f.close()
        mpi.world.barrier()

    def overlap(self, ov_nn, other):
        """Matrix element overlap determined from pair density overlaps.

        Parameters
        ----------
        ov_nn: array
            Wave function overlap factors from a displaced calculator.

            Index 0 corresponds to our own wavefunctions and
            index 1 to the others wavefunctions

        Returns
        -------
        ov_pp: array
            Overlap
        """
        # XXX ov_pp = self.kss.overlap(ov_nn, other.kss)
        ov_pp = self.Om.kss.overlap(ov_nn, other.Om.kss)
        self.diagonalize()
        other.diagonalize()
        # ov[pLm, pLo] = Om[pLm, :pKm]* ov[:pKm, pLo]
        return np.dot(self.Om.eigenvectors.conj(),
                      # ov[pKm, pLo] = ov[pKm, :pKo] Om[pLo, :pKo].T
                      np.dot(ov_pp, other.Om.eigenvectors.T))

    def __getitem__(self, i):
        if not self.diagonalized:
            self.diagonalize()
        return list.__getitem__(self, i)

    def __iter__(self):
        if not self.diagonalized:
            self.diagonalize()
        return list.__iter__(self)

    def __len__(self):
        if not self.diagonalized:
            self.diagonalize()
        return list.__len__(self)

    def __del__(self):
        try:
            self.timer.write(self.log.fd)
        except AttributeError:
            pass


def d2Excdnsdnt(dup, ddn):
    """Second derivative of Exc polarised"""
    res = [[0, 0], [0, 0]]
    for ispin in range(2):
        for jspin in range(2):
            res[ispin][jspin] = np.zeros(dup.shape)
            _gpaw.d2Excdnsdnt(dup, ddn, ispin, jspin, res[ispin][jspin])
    return res


def d2Excdn2(den):
    """Second derivative of Exc unpolarised"""
    res = np.zeros(den.shape)
    _gpaw.d2Excdn2(den, res)
    return res


class LrTDDFTExcitation(Excitation):

    def __init__(self, Om=None, i=None,
                 e=None, m=None, string=None):

        # multiplicity comes from Kohn-Sham contributions
        self.fij = 1

        if string is not None:
            self.fromstring(string)
            return None

        # define from the diagonalized Omega matrix
        if Om is not None:
            if i is None:
                raise RuntimeError

            ev = Om.eigenvalues[i]
            if ev < 0:
                # we reached an instability, mark it with a negative value
                self.energy = -sqrt(-ev)
            else:
                self.energy = sqrt(ev)
            self.f = Om.eigenvectors[i]
            self.kss = Om.kss

            self.kss.set_arrays()
            self.me = np.dot(self.f, self.kss.me)
            erat_k = np.sqrt(self.kss.energies / self.energy)
            wght_k = np.sqrt(self.kss.fij) * self.f
            ew_k = erat_k * wght_k
            self.mur = np.dot(ew_k, self.kss.mur)
            if self.kss.muv is not None:
                self.muv = np.dot(ew_k, self.kss.muv)
            else:
                self.muv = None
            if self.kss.magn is not None:
                self.magn = np.dot(wght_k / erat_k, self.kss.magn)
            else:
                self.magn = None

            return

        # define from energy and matrix element
        if e is not None:
            self.energy = e
            if m is None:
                raise RuntimeError
            self.me = m
            return

        raise RuntimeError

    def density_change(self, paw):
        """get the density change associated with this transition"""
        raise NotImplementedError

    def fromstring(self, string):
        l = string.split()
        self.energy = float(l.pop(0))
        if len(l) == 3:  # old writing style
            self.me = np.array([float(l.pop(0)) for i in range(3)])
        else:
            self.mur = np.array([float(l.pop(0)) for i in range(3)])
            self.me = - self.mur * sqrt(self.energy)
            self.muv = np.array([float(l.pop(0)) for i in range(3)])
            self.magn = np.array([float(l.pop(0)) for i in range(3)])

    def outstring(self):
        str = '%g ' % self.energy
        str += '  '
        for m in self.mur:
            str += '%12.4e' % m
        str += '  '
        for m in self.muv:
            str += '%12.4e' % m
        str += '  '
        for m in self.magn:
            str += '%12.4e' % m
        str += '\n'
        return str

    def __str__(self):
        m2 = np.sum(self.me * self.me)
        m = sqrt(m2)
        if m > 0:
            me = self.me / m
        else:
            me = self.me
        str = '<LrTDDFTExcitation> om=%g[eV] |me|=%g (%.2f,%.2f,%.2f)' % \
              (self.energy * Hartree, m, me[0], me[1], me[2])
        return str

    def analyse(self, min=.1):
        """Return an analysis string of the excitation"""
        osc = self.get_oscillator_strength()
        s = ('E=%.3f' % (self.energy * Hartree) + ' eV, ' +
             'f=%.5g' % osc[0] + ', (%.5g,%.5g,%.5g) ' %
             (osc[1], osc[2], osc[3]) + '\n')
        # 'R=%.5g' % self.get_rotatory_strength() + ' cgs\n')

        def sqr(x):
            return x * x
        spin = ['u', 'd']
        min2 = sqr(min)
        rest = np.sum(self.f**2)
        for f, k in zip(self.f, self.kss):
            f2 = sqr(f)
            if f2 > min2:
                s += '  %d->%d ' % (k.i, k.j) + spin[k.pspin] + ' '
                s += '%.3g \n' % f2
                rest -= f2
        s += '  rest=%.3g' % rest
        return s


def photoabsorption_spectrum(excitation_list, spectrum_file=None,
                             e_min=None, e_max=None, delta_e=None,
                             energyunit='eV',
                             folding='Gauss', width=0.1, comment=None):
    """Uniform absorption spectrum interface

    Parameters:
    ================= ===================================================
    ``exlist``        ExcitationList
    ``spectrum_file`` File name for the output file, STDOUT if not given
    ``e_min``         min. energy, set to cover all energies if not given
    ``e_max``         max. energy, set to cover all energies if not given
    ``delta_e``       energy spacing
    ``energyunit``    Energy unit, default 'eV'
    ``folding``       Gauss (default) or Lorentz
    ``width``         folding width in terms of the chosen energyunit
    ================= ===================================================
    all energies in [eV]
    """

    spectrum(exlist=excitation_list, filename=spectrum_file,
             emin=e_min, emax=e_max,
             de=delta_e, energyunit=energyunit,
             folding=folding, width=width,
             comment=comment)



import numpy as np
from ase.units import kB, Hartree, Bohr
from ase.data.vdw import vdw_radii

from gpaw.solvation.gridmem import NeedsGD
from gpaw.fd_operators import Gradient
from gpaw.io.logger import indent


BAD_RADIUS_MESSAGE = "All atomic radii have to be finite and >= zero."


def set_log_and_check_radii(obj, atoms, log):
    radii = np.array(obj.atomic_radii(atoms), dtype=float)
    obj.atomic_radii_output = radii
    obj.symbols = atoms.get_chemical_symbols()
    log('  Atomic radii for %s:' % (obj.__class__, ))
    for a, (s, r) in enumerate(
            zip(obj.symbols, radii)):
        log('    %3d %-2s %10.5f' % (a, s, r))
    if not np.isfinite(radii).all() or (radii < 0).any():
        raise ValueError(BAD_RADIUS_MESSAGE)


def get_pbc_positions(atoms, r_max):
    """Return dict mapping atom index to positions in Bohr.

    With periodic boundary conditions, it also includes neighbouring
    cells up to a distance of r_max (in Bohr).
    """
    # code snippet taken from ase/calculators/vdwcorrection.py
    pbc_c = atoms.get_pbc()
    cell_cv = atoms.get_cell() / Bohr
    Rcell_c = np.sqrt(np.sum(cell_cv ** 2, axis=1))
    ncells_c = np.ceil(np.where(pbc_c, 1. + r_max / Rcell_c, 1)).astype(int)
    pos_aav = {}
    # loop over all atoms in the cell (and neighbour cells for PBC)
    for index1, atom in enumerate(atoms):
        pos = atom.position / Bohr
        pos_aav[index1] = np.empty((np.prod(ncells_c * 2 - 1), 3))
        # loops over neighbour cells
        index2 = 0
        for ix in range(-ncells_c[0] + 1, ncells_c[0]):
            for iy in range(-ncells_c[1] + 1, ncells_c[1]):
                for iz in range(-ncells_c[2] + 1, ncells_c[2]):
                    i_c = np.array([ix, iy, iz])
                    pos_aav[index1][index2, :] = pos + np.dot(i_c, cell_cv)
                    index2 += 1
    return pos_aav


class Cavity(NeedsGD):
    """Base class for representing a cavity in the solvent.

    Attributes:
    g_g           -- The cavity on the fine grid. It varies from zero at the
                     solute location to one in the bulk solvent.
    del_g_del_n_g -- The partial derivative of the cavity with respect to
                     the electron density on the fine grid.
    grad_g_vg     -- The gradient of the cavity on the fine grid.
    V             -- global Volume in Bohr ** 3 or None
    A             -- global Area in Bohr ** 2 or None
    """

    def __init__(self, surface_calculator=None, volume_calculator=None):
        """Constructor for the Cavity class.

        Arguments:
        surface_calculator -- A SurfaceCalculator instance or None
        volume_calculator  -- A VolumeCalculator instance or None
        """
        NeedsGD.__init__(self)
        self.g_g = None
        self.del_g_del_n_g = None
        self.grad_g_vg = None
        self.surface_calculator = surface_calculator
        self.volume_calculator = volume_calculator
        self.V = None  # global Volume
        self.A = None  # global Surface

    def write(self, writer):
        pass

    def read(self, reader):
        pass

    def estimate_memory(self, mem):
        ngrids = 1 + self.depends_on_el_density
        mem.subnode('Distribution Function', ngrids * self.gd.bytecount())
        mem.subnode(
            'Gradient of Distribution Function', 3 * self.gd.bytecount()
        )
        if self.surface_calculator is not None:
            self.surface_calculator.estimate_memory(
                mem.subnode('Surface Calculator')
            )
        if self.volume_calculator is not None:
            self.volume_calculator.estimate_memory(
                mem.subnode('Volume Calculator')
            )

    def update(self, atoms, density):
        """Update the cavity and its gradient.

        atoms are None, iff they have not changed.

        Return whether the cavity has changed.
        """
        raise NotImplementedError()

    def update_vol_surf(self):
        """Update volume and surface."""
        if self.surface_calculator is not None:
            self.surface_calculator.update(self)
        if self.volume_calculator is not None:
            self.volume_calculator.update(self)

    def communicate_vol_surf(self, world):
        """Communicate global volume and surface."""
        if self.surface_calculator is not None:
            A = np.array([self.surface_calculator.A])
            self.gd.comm.sum(A)
            world.broadcast(A, 0)
            self.A = A[0]
        else:
            self.A = None
        if self.volume_calculator is not None:
            V = np.array([self.volume_calculator.V])
            self.gd.comm.sum(V)
            world.broadcast(V, 0)
            self.V = V[0]
        else:
            self.V = None

    def allocate(self):
        NeedsGD.allocate(self)
        self.g_g = self.gd.empty()
        self.grad_g_vg = self.gd.empty(3)
        if self.depends_on_el_density:
            self.del_g_del_n_g = self.gd.empty()
        if self.surface_calculator is not None:
            self.surface_calculator.allocate()
        if self.volume_calculator is not None:
            self.volume_calculator.allocate()

    def set_grid_descriptor(self, gd):
        NeedsGD.set_grid_descriptor(self, gd)
        if self.surface_calculator is not None:
            self.surface_calculator.set_grid_descriptor(gd)
        if self.volume_calculator is not None:
            self.volume_calculator.set_grid_descriptor(gd)

    def get_del_r_vg(self, atom_index, density):
        """Return spatial derivatives with respect to atomic position."""
        raise NotImplementedError()

    @property
    def depends_on_el_density(self):
        """Return whether the cavity depends on the electron density."""
        raise NotImplementedError()

    @property
    def depends_on_atomic_positions(self):
        """Return whether the cavity depends explicitly on atomic positions."""
        raise NotImplementedError()

    def __str__(self):
        s = 'Cavity: %s\n' % (self.__class__, )
        for calc, calcname in ((self.surface_calculator, 'Surface'),
                               (self.volume_calculator, 'Volume')):
            if calc is None:
                s += '  %s Calculator: None\n' % (calcname, )
            else:
                s += indent(str(calc))
        return s

    def update_atoms(self, atoms, log):
        """Inexpensive initialization when atoms change."""
        pass

    def summary(self, log):
        """Log cavity surface area and volume."""
        A = ('{:.5f}'.format(self.A * Bohr ** 2) if self.A is not None
             else 'not calculated (no calculator defined)')
        V = ('{:.5f}'.format(self.V * Bohr ** 3) if self.V is not None
             else 'not calculated (no calculator defined)')
        log('Solvation cavity surface area: %s' % (A, ))
        log('Solvation cavity volume: %s' % (V, ))


class EffectivePotentialCavity(Cavity):
    """Cavity built from effective potential and Boltzmann distribution.

    See also
    A. Held and M. Walter, J. Chem. Phys. 141, 174108 (2014).
    """

    def __init__(
        self,
        effective_potential,
        temperature,
        surface_calculator=None, volume_calculator=None
    ):
        """Constructor for the EffectivePotentialCavity class.

        Additional arguments not present in base Cavity class:
        effective_potential -- A Potential instance.
        temperature         -- Temperature for the Boltzmann distribution
                               in Kelvin.
        """
        Cavity.__init__(self, surface_calculator, volume_calculator)
        self.effective_potential = effective_potential
        self.temperature = float(temperature)
        self.minus_beta = -1. / (kB * temperature / Hartree)

    def write(self, writer):
        writer.write(effective_potential=self.effective_potential,
                     temperature=self.temperature,
                     surface_calculator=self.surface_calculator,
                     volume_calculator=self.volume_calculator)

    # For future, not used at the moment
    def read(self, reader):
        c = reader.parameters.cavity
        self.effective_potential = c.effective_potential
        self.temperature = c.temperature
        self.surface_calculator = c.surface_calculator
        self.volume_calculator = c.volume_calculator

    def estimate_memory(self, mem):
        Cavity.estimate_memory(self, mem)
        self.effective_potential.estimate_memory(
            mem.subnode('Effective Potential')
        )

    def set_grid_descriptor(self, gd):
        Cavity.set_grid_descriptor(self, gd)
        self.effective_potential.set_grid_descriptor(gd)

    def allocate(self):
        Cavity.allocate(self)
        self.effective_potential.allocate()

    def update(self, atoms, density):
        if not self.effective_potential.update(atoms, density):
            return False
        u_g = self.effective_potential.u_g
        np.exp(u_g * self.minus_beta, out=self.g_g)
        if self.depends_on_el_density:
            self.del_g_del_n_g.fill(self.minus_beta)
            self.del_g_del_n_g *= self.g_g
            self.del_g_del_n_g *= self.effective_potential.del_u_del_n_g
        # This class supports the GradientSurface API,
        # so we can use it for the gradient also
        grad_inner_vg = self.get_grad_inner()
        del_outer_del_inner = self.get_del_outer_del_inner()
        for i in (0, 1, 2):
            np.multiply(
                grad_inner_vg[i],
                del_outer_del_inner,
                self.grad_g_vg[i]
            )
        return True

    def get_del_r_vg(self, atom_index, density):
        u = self.effective_potential
        del_u_del_r_vg = u.get_del_r_vg(atom_index, density)
        # there should be no more NaNs now, but let's keep the hint
        # asserts lim_(||r - r_atom|| -> 0) dg/du * du/dr_atom = 0
        # del_u_del_r_vg[np.isnan(del_u_del_r_vg)] = .0
        return self.minus_beta * self.g_g * del_u_del_r_vg

    @property
    def depends_on_el_density(self):
        return self.effective_potential.depends_on_el_density

    @property
    def depends_on_atomic_positions(self):
        return self.effective_potential.depends_on_atomic_positions

    def __str__(self):
        s = Cavity.__str__(self)
        s += indent(str(self.effective_potential))
        s += '  temperature: %s\n' % (self.temperature, )
        return s

    def update_atoms(self, atoms, log):
        self.effective_potential.update_atoms(atoms, log)

    # --- BEGIN GradientSurface API ---

    def get_grad_inner(self):
        return self.effective_potential.grad_u_vg

    def get_del_outer_del_inner(self):
        return self.minus_beta * self.g_g

    # --- END GradientSurface API ---


class Potential(NeedsGD):
    """Base class for describing an effective potential.

    Attributes:
    u_g           -- The potential on the fine grid in Hartree.
    grad_u_vg     -- The gradient of the potential of the fine grid.
    del_u_del_n_g -- Partial derivative with respect to the electron density.
    """

    def __init__(self):
        NeedsGD.__init__(self)
        self.u_g = None
        self.del_u_del_n_g = None
        self.grad_u_vg = None

    @property
    def depends_on_el_density(self):
        """Return whether the cavity depends on the electron density."""
        raise NotImplementedError()

    @property
    def depends_on_atomic_positions(self):
        """returns whether the cavity depends explicitly on atomic positions"""
        raise NotImplementedError()

    def estimate_memory(self, mem):
        ngrids = 1 + self.depends_on_el_density
        mem.subnode('Potential', ngrids * self.gd.bytecount())
        mem.subnode('Gradient of Potential', 3 * self.gd.bytecount())

    def allocate(self):
        NeedsGD.allocate(self)
        self.u_g = self.gd.empty()
        if self.depends_on_el_density:
            self.del_u_del_n_g = self.gd.empty()
        self.grad_u_vg = self.gd.empty(3)

    def update(self, atoms, density):
        """Update the potential and its gradient.

        atoms are None, iff they have not changed.

        Return whether the potential has changed.
        """
        raise NotImplementedError()

    def get_del_r_vg(self, atom_index, density):
        """Return spatial derivatives with respect to atomic position."""
        raise NotImplementedError()

    def __str__(self):
        return 'Potential: %s\n' % (self.__class__, )

    def update_atoms(self, atoms, log):
        """Inexpensive initialization when atoms change."""
        pass


def get_vdw_radii(atoms):
    """Returns a list of van der Waals radii for a given atoms object."""
    return [vdw_radii[n] for n in atoms.numbers]


class Power12Potential(Potential):
    """Inverse power law potential.

    A 1 / r ** 12 repulsive potential taking the value u0 at the atomic radius.

    See also
    A. Held and M. Walter, J. Chem. Phys. 141, 174108 (2014).

    Parameters:

    atomic_radii: function
        Callable mapping an ase.Atoms object to an iterable of atomic radii
        in Angstroms. If not provided, defaults to van der Waals radii.
    u0: float
        Strength of the potential at the atomic radius in eV.
        Defaults to 0.18 eV, the best-fit value for water from Held &
        Walter.
    pbc_cutoff: float
        Cutoff in eV for including neighbor cells in a calculation with
        periodic boundary conditions.
    """
    depends_on_el_density = False
    depends_on_atomic_positions = True

    def __init__(self, atomic_radii=None, u0=0.180, pbc_cutoff=1e-6,
                 tiny=1e-10):
        Potential.__init__(self)
        if atomic_radii is None:
            atomic_radii = get_vdw_radii
        self.atomic_radii = atomic_radii
        self.u0 = float(u0)
        self.pbc_cutoff = float(pbc_cutoff)
        self.tiny = float(tiny)
        self.r12_a = None
        self.r_vg = None
        self.pos_aav = None
        self.del_u_del_r_vg = None
        self.atomic_radii_output = None
        self.symbols = None

    def estimate_memory(self, mem):
        Potential.estimate_memory(self, mem)
        nbytes = self.gd.bytecount()
        mem.subnode('Coordinates', 3 * nbytes)
        mem.subnode('Atomic Position Derivative', 3 * nbytes)

    def allocate(self):
        Potential.allocate(self)
        self.r_vg = self.gd.get_grid_point_coordinates()
        self.del_u_del_r_vg = self.gd.empty(3)

    def update(self, atoms, density):
        if atoms is None:
            return False
        self.r12_a = (self.atomic_radii_output / Bohr) ** 12
        r_cutoff = (self.r12_a.max() * self.u0 / self.pbc_cutoff) ** (1. / 12.)
        self.pos_aav = get_pbc_positions(atoms, r_cutoff)
        self.u_g.fill(.0)
        self.grad_u_vg.fill(.0)
        na = np.newaxis
        for index, pos_av in self.pos_aav.items():
            r12 = self.r12_a[index]
            for pos_v in pos_av:
                origin_vg = pos_v[:, na, na, na]
                r_diff_vg = self.r_vg - origin_vg
                r_diff2_g = (r_diff_vg ** 2).sum(0)
                r_diff2_g[r_diff2_g < self.tiny] = self.tiny
                u_g = r12 / r_diff2_g ** 6
                self.u_g += u_g
                u_g /= r_diff2_g
                r_diff_vg *= u_g[na, ...]
                self.grad_u_vg += r_diff_vg
        self.u_g *= self.u0 / Hartree
        self.grad_u_vg *= -12. * self.u0 / Hartree
        # avoid overflow in norm calculation:
        self.grad_u_vg[self.grad_u_vg < -1e20] = -1e20
        self.grad_u_vg[self.grad_u_vg > 1e20] = 1e20
        return True

    def get_del_r_vg(self, atom_index, density):
        u0 = self.u0 / Hartree
        r12 = self.r12_a[atom_index]
        na = np.newaxis
        self.del_u_del_r_vg.fill(.0)
        for pos_v in self.pos_aav[atom_index]:
            origin_vg = pos_v[:, na, na, na]
            diff_vg = self.r_vg - origin_vg
            diff2_g = (diff_vg ** 2).sum(0)
            diff2_g[diff2_g < self.tiny] = self.tiny
            diff2_g **= 7
            diff_vg /= diff2_g[na, ...]
            self.del_u_del_r_vg += diff_vg
        self.del_u_del_r_vg *= (12. * u0 * r12)
        return self.del_u_del_r_vg

    def __str__(self):
        s = Potential.__str__(self)
        s += '  atomic_radii: %s\n' % (self.atomic_radii, )
        s += '  u0: %s\n' % (self.u0, )
        s += '  pbc_cutoff: %s\n' % (self.pbc_cutoff, )
        s += '  tiny: %s\n' % (self.tiny, )
        return s

    def update_atoms(self, atoms, log):
        set_log_and_check_radii(self, atoms, log)

    def write(self, writer):
        writer.write(
            name=self.__class__.__name__,
            atomic_radii=self.atomic_radii_output,
            u0=self.u0,
            pbc_cutoff=self.pbc_cutoff,
            tiny=self.tiny)


class SmoothStepCavity(Cavity):
    """Base class for cavities based on a smooth step function and a density.

    Attributes:
    del_g_del_rho_g -- Partial derivative with respect to the density.
    """

    def __init__(
        self,
        density,
        surface_calculator=None, volume_calculator=None,
    ):
        """Constructor for the SmoothStepCavity class.

        Additional arguments not present in the base Cavity class:
        density -- A Density instance
        """
        Cavity.__init__(self, surface_calculator, volume_calculator)
        self.del_g_del_rho_g = None
        self.density = density

    @property
    def depends_on_el_density(self):
        return self.density.depends_on_el_density

    @property
    def depends_on_atomic_positions(self):
        return self.density.depends_on_atomic_positions

    def set_grid_descriptor(self, gd):
        Cavity.set_grid_descriptor(self, gd)
        self.density.set_grid_descriptor(gd)

    def estimate_memory(self, mem):
        Cavity.estimate_memory(self, mem)
        mem.subnode('Cavity Derivative', self.gd.bytecount())
        self.density.estimate_memory(mem.subnode('Density'))

    def allocate(self):
        Cavity.allocate(self)
        self.del_g_del_rho_g = self.gd.empty()
        self.density.allocate()

    def update(self, atoms, density):
        if not self.density.update(atoms, density):
            return False
        self.update_smooth_step(self.density.rho_g)
        if self.depends_on_el_density:
            np.multiply(
                self.del_g_del_rho_g,
                self.density.del_rho_del_n_g,
                self.del_g_del_n_g
            )
        # This class supports the GradientSurface API,
        # so we can use it for the gradient also
        grad_inner_vg = self.get_grad_inner()
        del_outer_del_inner = self.get_del_outer_del_inner()
        for i in (0, 1, 2):
            np.multiply(
                grad_inner_vg[i],
                del_outer_del_inner,
                self.grad_g_vg[i])
        return True

    def update_smooth_step(self, rho_g):
        """Calculate self.g_g and self.del_g_del_rho_g."""
        raise NotImplementedError()

    def get_del_r_vg(self, atom_index, density):
        return self.del_g_del_rho_g * self.density.get_del_r_vg(
            atom_index, density
        )

    def __str__(self):
        s = Cavity.__str__(self)
        s += indent(str(self.density))
        return s

    def update_atoms(self, atoms, log):
        self.density.update_atoms(atoms, log)

    # --- BEGIN GradientSurface API ---
    def get_grad_inner(self):
        return self.density.grad_rho_vg

    def get_del_outer_del_inner(self):
        return self.del_g_del_rho_g

    # --- END GradientSurface API ---


class Density(NeedsGD):
    """Class representing a density for the use with a SmoothStepCavity.

    Attributes:
    rho_g           -- The density on the fine grid in 1 / Bohr ** 3.
    del_rho_del_n_g -- Partial derivative with respect to the electron density.
    grad_rho_vg     -- The gradient of the density on the fine grid.
    """

    def __init__(self):
        NeedsGD.__init__(self)
        self.rho_g = None
        self.del_rho_del_n_g = None
        self.grad_rho_vg = None

    def estimate_memory(self, mem):
        nbytes = self.gd.bytecount()
        mem.subnode('Density', nbytes)
        if self.depends_on_el_density:
            mem.subnode('Density Derivative', nbytes)
        mem.subnode('Gradient of Density', 3 * nbytes)

    def allocate(self):
        NeedsGD.allocate(self)
        self.rho_g = self.gd.empty()
        if self.depends_on_el_density:
            self.del_rho_del_n_g = self.gd.empty()
        self.grad_rho_vg = self.gd.empty(3)

    def update(self, atoms, density):
        raise NotImplementedError()

    @property
    def depends_on_el_density(self):
        raise NotImplementedError()

    @property
    def depends_on_atomic_positions(self):
        raise NotImplementedError()

    def __str__(self):
        return "Density: %s\n" % (self.__class__, )

    def update_atoms(self, atoms, log):
        """Inexpensive initialization when atoms change."""
        pass


class FDGradientDensity(Density):
    """Base class for all Density classes with finite difference gradient"""

    def __init__(self, boundary_value, nn=3):
        """Constructur for the FDGradientDensity class.

        Arguments:
        boundary_value -- Boundary value of rho_g
                          (for non-periodic directions).
        nn             -- Stencil size for the finite difference gradient.
        """
        Density.__init__(self)
        self.boundary_value = boundary_value
        self.nn = nn
        self.gradient = None

    def allocate(self):
        Density.allocate(self)
        self.gradient = [
            Gradient(self.gd, i, 1.0, self.nn) for i in (0, 1, 2)
        ]

    def update(self, atoms, density):
        changed = self.update_only_density(atoms, density)
        if changed:
            self.update_gradient()
        return changed

    def update_gradient(self):
        if self.boundary_value != 0:
            in_g = self.rho_g - self.boundary_value
        else:
            in_g = self.rho_g
        for i in (0, 1, 2):
            self.gradient[i].apply(in_g, self.grad_rho_vg[i])


class ElDensity(FDGradientDensity):
    """Wrapper class for using the electron density in a SmoothStepCavity.

    (Hopefully small) negative values of the electron density are set to zero.
    """

    depends_on_el_density = True
    depends_on_atomic_positions = False

    def __init__(self, nn=3):
        """Constructor for the ElDensity class.

        Arguments:
        nn -- Stencil size for the finite difference gradient.
        """
        FDGradientDensity.__init__(self, boundary_value=.0, nn=nn)

    def allocate(self):
        FDGradientDensity.allocate(self)
        self.del_rho_del_n_g = 1.  # free array

    def update_only_density(self, atoms, density):
        self.rho_g[:] = density.nt_g
        self.rho_g[self.rho_g < .0] = .0
        return True


# TODO: implement analytic gradient for SSS09Density
class SSS09Density(FDGradientDensity):
    """Fake density from atomic radii for the use in a SmoothStepCavity.

    Following V. M. Sanchez, M. Sued and D. A. Scherlis,
    J. Chem. Phys. 131, 174108 (2009).
    """

    depends_on_el_density = False
    depends_on_atomic_positions = True

    def __init__(self, atomic_radii, pbc_cutoff=1e-3, nn=3, tiny=1e-100):
        """Constructor for the SSS09Density class.

        Arguments:
        atomic_radii -- Callable mapping an ase.Atoms object
                        to an iterable of atomic radii in Angstroms.
        pbc_cutoff   -- Cutoff in eV for including neighbor cells in
                        a calculation with periodic boundary conditions.
        nn           -- Stencil size for the finite difference gradient.
        """
        FDGradientDensity.__init__(self, boundary_value=.0, nn=nn)
        self.atomic_radii = atomic_radii
        self.atomic_radii_output = None
        self.symbols = None
        self.pbc_cutoff = float(pbc_cutoff)
        self.tiny = float(tiny)
        self.pos_aav = None
        self.r_vg = None
        self.del_rho_del_r_vg = None

    def estimate_memory(self, mem):
        FDGradientDensity.estimate_memory(self, mem)
        nbytes = self.gd.bytecount()
        mem.subnode('Coordinates', 3 * nbytes)
        mem.subnode('Atomic Position Derivative', 3 * nbytes)

    def allocate(self):
        FDGradientDensity.allocate(self)
        self.r_vg = self.gd.get_grid_point_coordinates()
        self.del_rho_del_r_vg = self.gd.empty(3)

    def update_only_density(self, atoms, density):
        if atoms is None:
            return False
        r_a = self.atomic_radii_output / Bohr
        r_cutoff = r_a.max() - np.log(self.pbc_cutoff)
        self.pos_aav = get_pbc_positions(atoms, r_cutoff)
        self.rho_g.fill(.0)
        na = np.newaxis
        for index, pos_av in self.pos_aav.items():
            for pos_v in pos_av:
                origin_vg = pos_v[:, na, na, na]
                r_diff_vg = self.r_vg - origin_vg
                norm_r_diff_g = (r_diff_vg ** 2).sum(0) ** .5
                self.rho_g += np.exp(r_a[index] - norm_r_diff_g)
        return True

    def get_del_r_vg(self, atom_index, density):
        r_a = self.atomic_radii_output[atom_index] / Bohr
        na = np.newaxis
        self.del_rho_del_r_vg.fill(.0)
        for pos_v in self.pos_aav[atom_index]:
            origin_vg = pos_v[:, na, na, na]
            r_diff_vg = self.r_vg - origin_vg
            norm_r_diff_g = np.sum(r_diff_vg ** 2, axis=0) ** .5
            norm_r_diff_g[norm_r_diff_g < self.tiny] = self.tiny
            exponential = np.exp(r_a - norm_r_diff_g)
            exponential /= norm_r_diff_g
            r_diff_vg *= exponential[na, ...]
            self.del_rho_del_r_vg += r_diff_vg
        return self.del_rho_del_r_vg

    def __str__(self):
        s = FDGradientDensity.__str__(self)
        s += '  atomic_radii: %s\n' % (self.atomic_radii, )
        s += '  pbc_cutoff: %s\n' % (self.pbc_cutoff, )
        s += '  tiny: %s\n' % (self.tiny, )
        return s

    def update_atoms(self, atoms, log):
        set_log_and_check_radii(self, atoms, log)


class ADM12SmoothStepCavity(SmoothStepCavity):
    """Cavity from smooth step function.

    Following O. Andreussi, I. Dabo, and N. Marzari,
    J. Chem. Phys. 136, 064102 (2012).
    """

    def __init__(
        self,
        rhomin, rhomax, epsinf,
        density,
        surface_calculator=None, volume_calculator=None
    ):
        """Constructor for the ADM12SmoothStepCavity class.

        Additional arguments not present in the SmoothStepCavity class:
        rhomin -- Lower density isovalue in 1 / Angstrom ** 3.
        rhomax -- Upper density isovalue in 1 / Angstrom ** 3.
        epsinf -- Static dielectric constant of the solvent.
        """
        SmoothStepCavity.__init__(
            self, density, surface_calculator, volume_calculator
        )
        self.rhomin = float(rhomin)
        self.rhomax = float(rhomax)
        self.epsinf = float(epsinf)

    def update_smooth_step(self, rho_g):
        eps = self.epsinf
        inside = rho_g > self.rhomax * Bohr ** 3
        outside = rho_g < self.rhomin * Bohr ** 3
        transition = np.logical_not(
            np.logical_or(inside, outside)
        )
        self.g_g[inside] = .0
        self.g_g[outside] = 1.
        self.del_g_del_rho_g.fill(.0)
        t, dt = self._get_t_dt(np.log(rho_g[transition]))
        if eps == 1.0:
            # lim_{eps -> 1} (eps - eps ** t) / (eps - 1) = 1 - t
            self.g_g[transition] = t
            self.del_g_del_rho_g[transition] = dt / rho_g[transition]
        else:
            eps_to_t = eps ** t
            self.g_g[transition] = (eps_to_t - 1.) / (eps - 1.)
            self.del_g_del_rho_g[transition] = (
                eps_to_t * np.log(eps) * dt
            ) / (
                rho_g[transition] * (eps - 1.)
            )

    def _get_t_dt(self, x):
        lnmax = np.log(self.rhomax * Bohr ** 3)
        lnmin = np.log(self.rhomin * Bohr ** 3)
        twopi = 2. * np.pi
        arg = twopi * (lnmax - x) / (lnmax - lnmin)
        t = (arg - np.sin(arg)) / twopi
        dt = -2. * np.sin(arg / 2.) ** 2 / (lnmax - lnmin)
        return (t, dt)

    def __str__(self):
        s = SmoothStepCavity.__str__(self)
        s += '  rhomin: %s\n' % (self.rhomin, )
        s += '  rhomax: %s\n' % (self.rhomax, )
        s += '  epsinf: %s\n' % (self.epsinf, )
        return s


class FG02SmoothStepCavity(SmoothStepCavity):
    """Cavity from smooth step function.

    Following J. Fattebert and F. Gygi, J. Comput. Chem. 23, 662 (2002).
    """

    def __init__(
        self,
        rho0,
        beta,
        density,
        surface_calculator=None, volume_calculator=None
    ):
        """Constructor for the FG02SmoothStepCavity class.

        Additional arguments not present in the SmoothStepCavity class:
        rho0 -- Density isovalue in 1 / Angstrom ** 3.
        beta -- Parameter controlling the steepness of the transition.
        """
        SmoothStepCavity.__init__(
            self, density, surface_calculator, volume_calculator
        )
        self.rho0 = float(rho0)
        self.beta = float(beta)

    def update_smooth_step(self, rho_g):
        rho0 = self.rho0 / (1. / Bohr ** 3)
        rho_scaled_g = rho_g / rho0
        exponent = 2. * self.beta
        np.divide(1., rho_scaled_g ** exponent + 1., self.g_g)
        np.multiply(
            (-exponent / rho0) * rho_scaled_g ** (exponent - 1.),
            self.g_g ** 2,
            self.del_g_del_rho_g
        )

    def __str__(self):
        s = SmoothStepCavity.__str__(self)
        s += '  rho0: %s\n' % (self.rho0, )
        s += '  beta: %s\n' % (self.beta, )
        return s


class SurfaceCalculator(NeedsGD):
    """Base class for surface calculators.

    Attributes:
    A                 -- Local area in Bohr ** 2.
    delta_A_delta_g_g -- Functional derivative with respect to the cavity
                         on the fine grid.
    """

    def __init__(self):
        NeedsGD.__init__(self)
        self.A = None
        self.delta_A_delta_g_g = None

    def write(self, writer):
        pass

    def read(self, reader):
        pass

    def estimate_memory(self, mem):
        mem.subnode('Functional Derivative', self.gd.bytecount())

    def allocate(self):
        NeedsGD.allocate(self)
        self.delta_A_delta_g_g = self.gd.empty()

    def __str__(self):
        return 'Surface Calculator: %s\n' % (self.__class__, )

    def update(self, cavity):
        """Calculate A and delta_A_delta_g_g."""
        raise NotImplementedError()


class GradientSurface(SurfaceCalculator):
    """Class for getting the surface area from the gradient of the cavity.

    See also W. Im, D. Beglov and B. Roux,
    Comput. Phys. Commun. 111, 59 (1998)
    and
    A. Held and M. Walter, J. Chem. Phys. 141, 174108 (2014).
    """

    def __init__(self, nn=3):
        SurfaceCalculator.__init__(self)
        self.nn = nn
        self.gradient = None
        self.gradient_out = None
        self.norm_grad_out = None
        self.div_tmp = None

    def write(self, writer):
        writer.write(
            name='GradientSurface',
            nn=self.nn)

    def read(self, reader):
        self.nn = reader.parameters.cavity.nn

    def estimate_memory(self, mem):
        SurfaceCalculator.estimate_memory(self, mem)
        nbytes = self.gd.bytecount()
        mem.subnode('Gradient', 4 * nbytes)
        mem.subnode('Divergence', nbytes)

    def allocate(self):
        SurfaceCalculator.allocate(self)
        self.gradient = [
            Gradient(self.gd, i, 1.0, self.nn) for i in (0, 1, 2)
        ]
        self.gradient_out = self.gd.empty(3)
        self.norm_grad_out = self.gd.empty()
        self.div_tmp = self.gd.empty()

    def update(self, cavity):
        del_outer_del_inner = cavity.get_del_outer_del_inner()
        sign = np.sign(del_outer_del_inner.max() + del_outer_del_inner.min())
        self.gradient_out[...] = cavity.get_grad_inner()
        self.norm_grad_out = (self.gradient_out ** 2).sum(0) ** .5
        self.A = sign * self.gd.integrate(
            del_outer_del_inner * self.norm_grad_out, global_integral=False
        )
        mask = self.norm_grad_out > 1e-12  # avoid division by zero or overflow
        imask = np.logical_not(mask)
        masked_norm_grad = self.norm_grad_out[mask]
        for i in (0, 1, 2):
            self.gradient_out[i][mask] /= masked_norm_grad
            # set limit for later calculations:
            self.gradient_out[i][imask] = .0
        self.calc_div(self.gradient_out, self.delta_A_delta_g_g)
        if sign == 1:
            self.delta_A_delta_g_g *= -1.

    def calc_div(self, vec, out):
        self.gradient[0].apply(vec[0], out)
        self.gradient[1].apply(vec[1], self.div_tmp)
        out += self.div_tmp
        self.gradient[2].apply(vec[2], self.div_tmp)
        out += self.div_tmp


class VolumeCalculator(NeedsGD):
    """Base class for volume calculators.

    Attributes:
    V                 -- Local volume in Bohr ** 3.
    delta_V_delta_g_g -- Functional derivative with respect to the cavity
                         on the fine grid.
    """

    def __init__(self):
        NeedsGD.__init__(self)
        self.V = None
        self.delta_V_delta_g_g = None

    def estimate_memory(self, mem):
        mem.subnode('Functional Derivative', self.gd.bytecount())

    def allocate(self):
        NeedsGD.allocate(self)
        self.delta_V_delta_g_g = self.gd.empty()

    def __str__(self):
        return "Volume Calculator: %s\n" % (self.__class__, )

    def update(self, cavity):
        """Calculate V and delta_V_delta_g_g"""
        raise NotImplementedError()


class KB51Volume(VolumeCalculator):
    """KB51 Volume Calculator.

    V = Integral(1 - g) + kappa_T * k_B * T

    Following J. G. Kirkwood and F. P. Buff, J. Chem. Phys. 19, 6, 774 (1951).
    """

    def __init__(self, compressibility=.0, temperature=.0):
        """Constructor for KB51Volume class.

        Arguments:
        compressibility -- Isothermal compressibility of the solvent
                           in Angstrom ** 3 / eV.
        temperature     -- Temperature in Kelvin.
        """
        VolumeCalculator.__init__(self)
        self.compressibility = float(compressibility)
        self.temperature = float(temperature)

    def __str__(self):
        s = VolumeCalculator.__str__(self)
        s += '  compressibility: %s\n' % (self.compressibility, )
        s += '  temperature:     %s\n' % (self.temperature, )
        return s

    def allocate(self):
        VolumeCalculator.allocate(self)
        self.delta_V_delta_g_g = -1.  # frees array

    def update(self, cavity):
        self.V = self.gd.integrate(1. - cavity.g_g, global_integral=False)
        V_compress = self.compressibility * kB * self.temperature / Bohr ** 3
        self.V += V_compress / self.gd.comm.size



import numpy as np

import ase.units as units
from ase.parallel import parprint
from ase.phonons import Displacement
from ase.utils.filecache import MultiFileJSONCache

from gpaw.calculator import GPAW
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.lcao.tightbinding import TightBinding
from gpaw.utilities import unpack2
from gpaw.utilities.tools import tri2full


# TODO replace parprint with something nicer
class ElectronPhononCoupling(Displacement):
    """Class for calculating the electron-phonon coupling in an LCAO basis.

    The derivative of the effective potential wrt atomic displacements is
    obtained from a finite difference approximation to the derivative by doing
    a self-consistent calculation for atomic displacements in the +/-
    directions. These calculations are carried out in the ``run`` member
    function.

    The subsequent calculation of the coupling matrix in the basis of atomic
    orbitals (or Bloch-sums hereof for periodic systems) is handled by the
    ``calculate_matrix`` member function.

    """

    def __init__(self, atoms, calc=None, supercell=(1, 1, 1), name='elph',
                 delta=0.01, calculate_forces=False):
        """Initialize with base class args and kwargs.

        Parameters
        ----------
        atoms: Atoms
            The atoms to work on.
        calc: GPAW
            Calculator for the supercell finite displacement calculation.
        supercell: tuple, list
            Size of supercell given by the number of repetitions (l, m, n) of
            the small unit cell in each direction.
        name: str
            Name to use for files (default: 'elph').
        delta: float
            Magnitude of displacements.
        calculate_forces: bool
            If true, also calculate and store the dynamical matrix.
        """

        parprint("DEPRECATION WARNING: This module is deprecated.")

        # Init base class and make the center cell in the supercell the
        # reference cell
        Displacement.__init__(self, atoms, calc=calc, supercell=supercell,
                              name=name, delta=delta, center_refcell=True)

        self.calculate_forces = calculate_forces
        # LCAO calculator
        self.calc_lcao = None
        # Supercell matrix
        self.g_xsNNMM = None
        self.basis_info = None
        self.supercell_cache = None

    def calculate(self, atoms_N, disp):
        return self(atoms_N)

    def __call__(self, atoms_N):
        """Extract effective potential and projector coefficients."""

        # Do calculation
        atoms_N.get_potential_energy()

        # Calculate forces if desired
        if self.calculate_forces:
            forces = atoms_N.get_forces()
        else:
            forces = None

        # Get calculator
        calc = atoms_N.calc
        if not isinstance(calc, GPAW):
            calc = calc.dft  # unwrap DFTD3 wrapper

        # Effective potential (in Hartree) and projector coefficients
        Vt_sG = calc.hamiltonian.vt_sG
        Vt_sG = calc.wfs.gd.collect(Vt_sG, broadcast=True)
        dH_asp = calc.hamiltonian.dH_asp

        setups = calc.wfs.setups
        nspins = calc.wfs.nspins
        gd_comm = calc.wfs.gd.comm

        dH_all_asp = {}
        for a, setup in enumerate(setups):
            ni = setup.ni
            nii = ni * (ni + 1) // 2
            dH_tmp_sp = np.zeros((nspins, nii))
            if a in dH_asp:
                dH_tmp_sp[:] = dH_asp[a]
            gd_comm.sum(dH_tmp_sp)
            dH_all_asp[a] = dH_tmp_sp

        output = {'Vt_sG': Vt_sG, 'dH_all_asp': dH_all_asp}
        if forces is not None:
            output['forces'] = forces
        return output

    def set_lcao_calculator(self, calc):
        """Set LCAO calculator for the calculation of the supercell matrix."""
        assert calc.wfs.mode == 'lcao', 'LCAO mode required.'
        assert not calc.symmetry.point_group, \
            'Point group symmetry not supported'
        self.calc_lcao = calc

    def set_basis_info(self, *args):
        """Store lcao basis info for atoms in reference cell in attribute.

        Parameters
        ----------
        args: tuple
            If the LCAO calculator is not available (e.g. if the supercell is
            loaded from file), the ``load_supercell_matrix`` member function
            provides the required info as arguments.

        """
        assert len(args) in (0, 2)
        if len(args) == 0:
            calc = self.calc_lcao
            setups = calc.wfs.setups
            bfs = calc.wfs.basis_functions
            nao_a = [setups[a].nao for a in range(len(self.atoms))]
            M_a = [bfs.M_a[a] for a in range(len(self.atoms))]
        else:
            M_a = args[0]
            nao_a = args[1]

        self.basis_info = {'M_a': M_a, 'nao_a': nao_a}

    def _calculate_supercell_entry(self, a, v, V1t_sG, dH1_asp, wfs,
                                   include_pseudo):
        kpt_u = wfs.kpt_u
        setups = wfs.setups
        nao = setups.nao
        bfs = wfs.basis_functions
        dtype = wfs.dtype
        nspins = wfs.nspins

        # Equilibrium atomic Hamiltonian matrix (projector coefficients)
        dH_asp = self.cache['eq']['dH_all_asp']

        # For the contribution from the derivative of the projectors
        dP_aqvMi = wfs.manytci.P_aqMi(self.indices, derivative=True)
        dP_qvMi = dP_aqvMi[a]

        # Array for different k-point components
        g_sqMM = np.zeros((nspins, len(kpt_u) // nspins, nao, nao), dtype)

        # 1) Gradient of effective potential
        parprint("Starting gradient of effective potential")
        for kpt in kpt_u:
            # Matrix elements
            geff_MM = np.zeros((nao, nao), dtype)
            bfs.calculate_potential_matrix(V1t_sG[kpt.s], geff_MM, q=kpt.q)
            tri2full(geff_MM, 'L')
            # Insert in array
            g_sqMM[kpt.s, kpt.q] += geff_MM
        parprint("Finished gradient of effective potential")

        if include_pseudo:
            parprint("Starting gradient of pseudo part")
            # 2) Gradient of non-local part (projectors)
            # parprint("Starting gradient of dH^a")
            P_aqMi = wfs.P_aqMi
            # 2a) dH^a part has contributions from all other atoms
            for kpt in kpt_u:
                # Matrix elements
                gp_MM = np.zeros((nao, nao), dtype)
                for a_, dH1_sp in dH1_asp.items():
                    dH1_ii = unpack2(dH1_sp[kpt.s])
                    P_Mi = P_aqMi[a_][kpt.q]
                    gp_MM += np.dot(P_Mi, np.dot(dH1_ii, P_Mi.T.conjugate()))
                g_sqMM[kpt.s, kpt.q] += gp_MM
            # parprint("Finished gradient of dH^a")

            # parprint("Starting gradient of projectors")
            # 2b) dP^a part has only contributions from the same atoms
            dH_ii = unpack2(dH_asp[a][kpt.s])
            for kpt in kpt_u:
                # XXX Sort out the sign here; conclusion -> sign = +1 !
                P1HP_MM = +1 * np.dot(dP_qvMi[kpt.q][v], np.dot(dH_ii,
                                      P_aqMi[a][kpt.q].T.conjugate()))
                # Matrix elements
                gp_MM = P1HP_MM + P1HP_MM.T.conjugate()
                g_sqMM[kpt.s, kpt.q] += gp_MM
            # parprint("Finished gradient of projectors")
            parprint("Finished gradient of pseudo part")
        return g_sqMM

    def calculate_supercell_matrix(self, name='supercell', filter=None,
                                   include_pseudo=True):
        """Calculate matrix elements of the el-ph coupling in the LCAO basis.

        This function calculates the matrix elements between LCAOs and local
        atomic gradients of the effective potential. The matrix elements are
        calculated for the supercell used to obtain finite-difference
        approximations to the derivatives of the effective potential wrt to
        atomic displacements.

        Parameters
        ----------
        name: str
            User specified name of the generated JSON cache.
            Default is 'supercell'.
        filter: str
            Fourier filter atomic gradients of the effective potential. The
            specified components (``normal`` or ``umklapp``) are removed
            (default: None).
        include_pseudo: bool
            Include the contribution from the psedupotential in the atomic
            gradients. If ``False``, only the gradient of the effective
            potential is included (default: True).
        """

        assert self.calc_lcao is not None, "Set LCAO calculator"

        # JSON cache
        self.supercell_cache = MultiFileJSONCache(name)

        # Supercell atoms
        atoms_N = self.atoms * self.supercell

        # Initialize calculator if required and extract useful quantities
        calc = self.calc_lcao
        if (not hasattr(calc.wfs, 'S_qMM') or
            not hasattr(calc.wfs.basis_functions, 'M_a')):
            calc.initialize(atoms_N)
            calc.initialize_positions(atoms_N)
        self.set_basis_info()

        # Extract useful objects from the calculator
        wfs = calc.wfs
        gd = calc.wfs.gd
        kd = calc.wfs.kd
        setups = wfs.setups
        nao = setups.nao
        nspins = wfs.nspins
        # FIXME: Domain parallelisation broken
        assert gd.comm.size == 1

        # If gamma calculation, overlap with neighboring cell cannot be removed
        if kd.gamma:
            print("WARNING: Gamma-point calculation.")
        else:
            # Bloch to real-space converter
            tb = TightBinding(atoms_N, calc)

        parprint("Calculating supercell matrix")

        parprint("Calculating real-space gradients")
        # Calculate finite-difference gradients (in Hartree / Bohr)
        V1t_xsG, dH1_xasp = self.calculate_gradient()
        parprint("Finished real-space gradients")

        # Fourier filter the atomic gradients of the effective potential
        if filter is not None:
            parprint("Fourier filtering gradients")
            for s in range(nspins):
                self.fourier_filter(V1t_xsG[:, s], components=filter)
            parprint("Finished Fourier filtering")

        # Check that the grid is the same as in the calculator
        assert np.all(V1t_xsG.shape[-3:] == (gd.N_c + gd.pbc_c - 1)), \
            "Mismatch in grids."

        with self.supercell_cache.lock('basis') as handle:
            if handle is not None:
                handle.save(self.basis_info)

        # Calculate < i k | grad H | j k >, i.e. matrix elements in Bloch basis
        parprint("Calculating gradient of PAW Hamiltonian")

        # Do each cartesian component separately
        for i, a in enumerate(self.indices):
            for v in range(3):
                # Corresponding array index
                x = 3 * i + v

                # If exist already, don't recompute
                with self.supercell_cache.lock(str(x)) as handle:
                    if handle is None:
                        continue

                    # parprint("Atom ", i+1, "/", len(self.indices), " ,
                    # direction ", v)
                    parprint("%s-gradient of atom %u" %
                             (['x', 'y', 'z'][v], a))

                    g_sqMM = self._calculate_supercell_entry(a, v, V1t_xsG[x],
                                                             dH1_xasp[x], wfs,
                                                             include_pseudo)

                    # Extract R_c=(0, 0, 0) block by Fourier transforming
                    if kd.gamma or kd.N_c is None:
                        g_sMM = g_sqMM[:, 0]
                    else:
                        # Convert to array
                        g_sMM = []
                        for s in range(nspins):
                            g_MM = tb.bloch_to_real_space(g_sqMM[s],
                                                          R_c=(0, 0, 0))
                            g_sMM.append(g_MM[0])  # [0] because of above
                        g_sMM = np.array(g_sMM)

                    # Reshape to global unit cell indices
                    N = np.prod(self.supercell)
                    # Number of basis function in the primitive cell
                    assert (nao % N) == 0, "Alarm ...!"
                    nao_cell = nao // N
                    g_sNMNM = g_sMM.reshape((nspins, N, nao_cell, N, nao_cell))
                    g_sNNMM = g_sNMNM.swapaxes(2, 3).copy()
                    parprint("Finished supercell matrix")

                    handle.save(g_sNNMM)
                if x == 0:
                    with self.supercell_cache.lock('info') as handle:
                        if handle is not None:
                            handle.save([g_sNNMM.shape, g_sNNMM.dtype.name])
        parprint("Finished gradient of PAW Hamiltonian")

    def load_supercell_matrix(self, name='supercell'):
        """Load supercell matrix from cache.

        Parameters
        ----------
        name: str
            User specified name of the cache.
        """
        self.supercell_cache = MultiFileJSONCache(name)
        self.basis_info = self.supercell_cache['basis']
        [shape, dtype] = self.supercell_cache['info']

        nx = len(self.indices) * 3
        self.g_xsNNMM = np.empty([nx, ] + list(shape), dtype=dtype)
        for x in range(nx):
            self.g_xsNNMM[x] = self.supercell_cache[str(x)]

    def apply_cutoff(self, cutmax=None, cutmin=None):
        """Zero matrix element inside/beyond the specified cutoffs.

        This method is not tested.
        This method does not respect minimum image convention.

        Parameters
        ----------
        cutmax: float
            Zero matrix elements for basis functions with a distance to the
            atomic gradient that is larger than the cutoff.
        cutmin: float
            Zero matrix elements where both basis functions have distances to
            the atomic gradient that is smaller than the cutoff.
        """

        if cutmax is not None:
            cutmax = float(cutmax)
        if cutmin is not None:
            cutmin = float(cutmin)

        # Reference to supercell matrix attribute
        g_xsNNMM = self.g_xsNNMM

        # Number of atoms and primitive cells
        N_atoms = len(self.indices)
        N = np.prod(self.supercell)
        nao = g_xsNNMM.shape[-1]
        nspins = g_xsNNMM.shape[1]

        # Reshape array
        g_avsNNMM = g_xsNNMM.reshape(N_atoms, 3, nspins, N, N, nao, nao)

        # Make slices for orbitals on atoms
        M_a = self.basis_info['M_a']
        nao_a = self.basis_info['nao_a']
        slice_a = []
        for a in range(len(self.atoms)):
            start = M_a[a]
            stop = start + nao_a[a]
            s = slice(start, stop)
            slice_a.append(s)

        # Lattice vectors
        R_cN = self.compute_lattice_vectors()

        # Unit cell vectors
        cell_vc = self.atoms.cell.transpose()
        # Atomic positions in reference cell
        pos_av = self.atoms.get_positions()

        # Create a mask array to zero the relevant matrix elements
        if cutmin is not None:
            mask_avsNNMM = np.zeros(g_avsNNMM.shape, dtype=bool)

        # Zero elements where one of the basis orbitals has a distance to atoms
        # (atomic gradients) in the reference cell larger than the cutoff
        for n in range(N):
            # Lattice vector to cell
            R_v = np.dot(cell_vc, R_cN[:, n])
            # Atomic positions in cell
            posn_av = pos_av + R_v
            for i, a in enumerate(self.indices):
                # Atomic distances wrt to the position of the gradient
                dist_a = np.sqrt(np.sum((pos_av[a] - posn_av)**2, axis=-1))

                if cutmax is not None:
                    # Atoms indices where the distance is larger than the max
                    # cufoff
                    j_a = np.where(dist_a > cutmax)[0]
                    # Zero elements
                    for j in j_a:
                        g_avsNNMM[a, :, :, n, :, slice_a[j], :] = 0.0
                        g_avsNNMM[a, :, :, :, n, :, slice_a[j]] = 0.0

                if cutmin is not None:
                    # Atoms indices where the distance is larger than the min
                    # cufoff
                    j_a = np.where(dist_a > cutmin)[0]
                    # Update mask to keep elements where one LCAO is outside
                    # the min cutoff
                    for j in j_a:
                        mask_avsNNMM[a, :, :, n, :, slice_a[j], :] = True
                        mask_avsNNMM[a, :, :, :, n, :, slice_a[j]] = True

        # Zero elements where both LCAOs are located within the min cutoff
        if cutmin is not None:
            g_avsNNMM[~mask_avsNNMM] = 0.0

    def lcao_matrix(self, u_l, omega_l):
        """Calculate the el-ph coupling in the electronic LCAO basis.

        For now, only works for Gamma-point phonons.

        This method is not tested.

        Parameters
        ----------
        u_l: np.ndarray
            Mass-scaled polarization vectors (in units of 1 / sqrt(amu)) of the
            phonons.
        omega_l: np.ndarray
            Vibrational frequencies in eV.
        """

        # Supercell matrix (Hartree / Bohr)
        assert self.g_xsNNMM is not None, "Load supercell matrix."
        assert self.g_xsNNMM.shape[2:4] == (1, 1)
        g_xsMM = self.g_xsNNMM[:, :, 0, 0, :, :]
        # Number of atomic orbitals
        # nao = g_xMM.shape[-1]
        # Number of phonon modes
        nmodes = u_l.shape[0]

        #
        u_lx = u_l.reshape(nmodes, 3 * len(self.atoms))
        # np.dot uses second to last index of second array
        g_lsMM = np.dot(u_lx, g_xsMM.transpose(2, 0, 1, 3))

        # Multiply prefactor sqrt(hbar / 2 * M * omega) in units of Bohr
        amu = units._amu  # atomic mass unit
        me = units._me   # electron mass
        g_lsMM /= np.sqrt(2 * amu / me / units.Hartree *
                          omega_l[:, :, np.newaxis, np.newaxis])
        # Convert to eV
        g_lsMM *= units.Hartree

        return g_lsMM

    def bloch_matrix(self, kpts, qpts, c_kn, u_ql,
                     omega_ql=None, kpts_from=None, spin=0,
                     name='supercell'):
        r"""Calculate el-ph coupling in the Bloch basis for the electrons.

        This function calculates the electron-phonon coupling between the
        specified Bloch states, i.e.::

                      ______
            mnl      / hbar               ^
           g    =   /-------  < m k + q | e  . grad V  | n k >
            kq    \/ 2 M w                 ql        q
                          ql

        In case the ``omega_ql`` keyword argument is not given, the bare matrix
        element (in units of eV / Ang) without the sqrt prefactor is returned.

        Phonon frequencies and mode vectors must be given in
        ase units.

        Parameters
        ----------
        kpts: np.ndarray or tuple
            k-vectors of the Bloch states. When a tuple of integers is given, a
            Monkhorst-Pack grid with the specified number of k-points along the
            directions of the reciprocal lattice vectors is generated.
        qpts: np.ndarray or tuple
            q-vectors of the phonons.
        c_kn: np.ndarray
            Expansion coefficients for the Bloch states. The ordering must be
            the same as in the ``kpts`` argument.
        u_ql: np.ndarray
            Mass-scaled polarization vectors (in units of 1 / sqrt(amu)) of the
            phonons. Again, the ordering must be the same as in the
            corresponding ``qpts`` argument.
        omega_ql: np.ndarray
            Vibrational frequencies in eV.
        kpts_from: list[int] or int
            Calculate only the matrix element for the k-vectors specified by
            their index in the ``kpts`` argument (default: all).
        spin: int
            In case of spin-polarised system, define which spin to use
            (0 or 1).
        """
        assert len(c_kn.shape) == 3
        assert len(u_ql.shape) == 4
        if omega_ql is not None:
            assert np.all(u_ql.shape[:2] == omega_ql.shape[:2])

        # Translate k-points into 1. BZ (required by ``find_k_plus_q``` member
        # function of the ```KPointDescriptor``).
        if isinstance(kpts, np.ndarray):
            assert kpts.shape[1] == 3, "kpts_kc array must be given"
            # XXX This does not seem to cause problems!
            kpts -= kpts.round()

        # Use the KPointDescriptor to keep track of the k and q-vectors
        kd_kpts = KPointDescriptor(kpts)
        kd_qpts = KPointDescriptor(qpts)
        # Check that number of k- and q-points agree with the number of Bloch
        # functions and polarization vectors
        assert kd_kpts.nbzkpts == len(c_kn)
        assert kd_qpts.nbzkpts == len(u_ql)

        # Include all k-point per default
        if kpts_from is None:
            kpts_kc = kd_kpts.bzk_kc
            kpts_k = range(kd_kpts.nbzkpts)
        else:
            kpts_kc = kd_kpts.bzk_kc[kpts_from]
            if isinstance(kpts_from, int):
                kpts_k = list([kpts_from])
            else:
                kpts_k = list(kpts_from)

        # Number of phonon modes and electronic bands
        nmodes = u_ql.shape[1]
        nbands = c_kn.shape[1]
        # Number of atoms displacements and basis functions
        ndisp = np.prod(u_ql.shape[2:])
        assert ndisp == (3 * len(self.indices))
        nao = c_kn.shape[2]

        # Lattice vectors
        R_cN = self.compute_lattice_vectors()
        # Number of unit cell in supercell
        N = np.prod(self.supercell)

        # Allocate array for couplings
        g_qklnn = np.zeros((kd_qpts.nbzkpts, len(kpts_kc), nmodes,
                            nbands, nbands), dtype=complex)

        self.supercell_cache = MultiFileJSONCache(name)
        self.basis_info = self.supercell_cache['basis']

        parprint("Calculating coupling matrix elements")
        for q, q_c in enumerate(kd_qpts.bzk_kc):
            # Find indices of k+q for the k-points
            kplusq_k = kd_kpts.find_k_plus_q(q_c, kpts_k=kpts_k)

            # Here, ``i`` is counting from 0 and ``k`` is the global index of
            # the k-point
            for i, (k, k_c) in enumerate(zip(kpts_k, kpts_kc)):
                # Check the wave vectors (adapted to the ``KPointDescriptor``
                # class)
                kplusq_c = k_c + q_c
                kplusq_c -= kplusq_c.round()
                assert np.allclose(kplusq_c, kd_kpts.bzk_kc[kplusq_k[i]]), \
                    (i, k, k_c, q_c, kd_kpts.bzk_kc[kplusq_k[i]])

                # LCAO coefficient for Bloch states
                ck_nM = c_kn[k]
                ckplusq_nM = c_kn[kplusq_k[i]]
                # Mass scaled polarization vectors
                u_lx = u_ql[q].reshape(nmodes, ndisp)

                # Multiply phase factors
                g_lnn = np.zeros((nmodes, nbands, nbands), dtype=complex)
                for x in range(ndisp):
                    # Allocate array
                    g_MM = np.zeros((nao, nao), dtype=complex)
                    g_sNNMM = self.supercell_cache[str(x)]
                    assert nao == g_sNNMM.shape[-1]
                    for m in range(N):
                        for n in range(N):
                            phase = self._get_phase_factor(R_cN, m, n, k_c,
                                                           q_c)
                            # Sum contributions from different cells
                            g_MM += g_sNNMM[spin, m, n, :, :] * phase

                    g_nn = np.dot(ckplusq_nM.conj(), np.dot(g_MM, ck_nM.T))
                    # not sure if einsum is faster or slower
                    # g_nn = np.einsum('ij,jk,kl->il',ckplusq_nM.conj(),
                    # g_MM, ck_nM.T)
                    # g_lnn += np.outer(u_lx[:,x],g_nn).reshape(nmodes,
                    # nbands, nbands)
                    g_lnn += np.einsum('i,kl->ikl', u_lx[:, x], g_nn)

                # Insert value
                g_qklnn[q, i] = g_lnn

                # XXX Temp
                if np.all(q_c == 0.0):
                    # These should be real. Problem is... they are usually not
                    print(g_qklnn[q].imag.min(), g_qklnn[q].imag.max())

        parprint("Finished calculation of coupling matrix elements")

        # Return the bare matrix element if frequencies are not given
        if omega_ql is None:
            # Convert to eV / Ang
            g_qklnn *= units.Hartree / units.Bohr
        else:
            # Multiply prefactor sqrt(hbar / 2 * M * omega) in units of Bohr
            amu = units._amu  # atomic mass unit
            me = units._me   # electron mass
            g_qklnn /= np.sqrt(2 * amu / me / units.Hartree *
                               omega_ql[:, np.newaxis, :,
                                        np.newaxis, np.newaxis])
            # Convert to eV
            g_qklnn *= units.Hartree

        # Return couplings in eV (or eV / Ang)
        return g_qklnn

    def fourier_filter(self, V1t_xG, components='normal', criteria=0):
        """Fourier filter atomic gradients of the effective potential.

        This method is not tested.

        Parameters
        ----------
        V1t_xG: np.ndarray
            Array representation of atomic gradients of the effective potential
            in the supercell grid.
        components: str
            Fourier components to filter out (``normal`` or ``umklapp``).
        """
        import numpy.fft as fft
        import numpy.linalg as la
        assert components in ['normal', 'umklapp']
        # Grid shape
        shape = V1t_xG.shape[-3:]

        # Primitive unit cells in Bohr/Bohr^-1
        cell_cv = self.atoms.get_cell() / units.Bohr
        reci_vc = 2 * np.pi * la.inv(cell_cv)
        norm_c = np.sqrt(np.sum(reci_vc**2, axis=0))
        # Periodic BC array
        pbc_c = np.array(self.atoms.get_pbc(), dtype=bool)

        # Supercell atoms and cell
        atoms_N = self.atoms * self.supercell
        supercell_cv = atoms_N.get_cell() / units.Bohr

        # q-grid in units of the grid spacing (FFT ordering)
        q_cG = np.indices(shape).reshape(3, -1)
        q_c = np.array(shape)[:, np.newaxis]
        q_cG += q_c // 2
        q_cG %= q_c
        q_cG -= q_c // 2

        # Locate q-points inside the Brillouin zone
        if criteria == 0:
            # Works for all cases
            # Grid spacing in direction of reciprocal lattice vectors
            h_c = np.sqrt(np.sum((2 * np.pi * la.inv(supercell_cv))**2,
                                 axis=0))
            # XXX Why does a "*=" operation on q_cG not work here ??
            q1_cG = q_cG * h_c[:, np.newaxis] / (norm_c[:, np.newaxis] / 2)
            mask_G = np.ones(np.prod(shape), dtype=bool)
            for i, pbc in enumerate(pbc_c):
                if not pbc:
                    continue
                mask_G &= (-1. < q1_cG[i]) & (q1_cG[i] <= 1.)
        else:
            # 2D hexagonal lattice
            # Projection of q points onto the periodic directions. Only in
            # these directions do normal and umklapp processees make sense.
            q_vG = np.dot(q_cG[pbc_c].T,
                          2 * np.pi * la.inv(supercell_cv).T[pbc_c]).T.copy()
            # Parametrize the BZ boundary in terms of the angle theta
            theta_G = np.arctan2(q_vG[1], q_vG[0]) % (np.pi / 3)
            phi_G = np.pi / 6 - np.abs(theta_G)
            qmax_G = norm_c[0] / 2 / np.cos(phi_G)
            norm_G = np.sqrt(np.sum(q_vG**2, axis=0))
            # Includes point on BZ boundary with +1e-2
            mask_G = (norm_G <= qmax_G + 1e-2)

        if components != 'normal':
            mask_G = ~mask_G

        # Reshape to grid shape
        mask_G.shape = shape

        for V1t_G in V1t_xG:
            # Fourier transform atomic gradient
            V1tq_G = fft.fftn(V1t_G)
            # Zero normal/umklapp components
            V1tq_G[mask_G] = 0.0
            # Fourier transform back
            V1t_G[:] = fft.ifftn(V1tq_G).real

    def calculate_gradient(self):
        """Calculate gradient of effective potential and projector coefs.

        This function loads the generated pickle files and calculates
        finite-difference derivatives.

        """

        # Array and dict for finite difference derivatives
        V1t_xsG = []
        dH1_xasp = []

        x = 0
        for a in self.indices:
            for v in 'xyz':
                # Note: self.name currently ignored in ase.phonon
                # name = '%s.%d%s' % (self.name, a, v)
                name = '%d%s' % (a, v)
                # Potential and atomic density matrix for atomic displacement
                Vtm_sG = self.cache[name + '-']['Vt_sG']
                dHm_asp = self.cache[name + '-']['dH_all_asp']
                Vtp_sG = self.cache[name + '+']['Vt_sG']
                dHp_asp = self.cache[name + '+']['dH_all_asp']

                # FD derivatives in Hartree / Bohr
                V1t_sG = (Vtp_sG - Vtm_sG) / (2 * self.delta / units.Bohr)
                V1t_xsG.append(V1t_sG)

                dH1_asp = {}
                for atom in dHm_asp.keys():
                    dH1_asp[atom] = (dHp_asp[atom] - dHm_asp[atom]) / \
                                    (2 * self.delta / units.Bohr)
                dH1_xasp.append(dH1_asp)
                x += 1

        return np.array(V1t_xsG), dH1_xasp

    def _get_phase_factor(self, R_cN, m, n, k_c, q_c):
        Rm_c = R_cN[:, m]
        Rn_c = R_cN[:, n]
        phase = np.exp(2.j * np.pi * (np.dot(k_c, Rm_c - Rn_c)
                                      + np.dot(q_c, Rm_c)))
        return phase



from __future__ import annotations

import numpy as np

from ase.utils import lazyproperty

from gpaw.projections import Projections, serial_comm
from gpaw.response import ResponseGroundStateAdapter, ResponseContext, timer
from gpaw.response.symmetry import KPointFinder
from gpaw.response.pw_parallelization import Blocks1D


class IrreducibleKPoint:
    """Irreducible k-point data pertaining to a certain set of transitions."""

    def __init__(self, ik, eps_h, f_h, Ph, psit_hG, h_myt):
        """Construct the IrreducibleKPoint data object.

        The data is indexed by the composite band and spin index h = (n, s),
        which can be unfolded to the local transition index myt.
        """
        self.ik = ik             # Irreducible k-point index
        self.eps_h = eps_h       # Eigenvalues
        self.f_h = f_h           # Occupation numbers
        self.Ph = Ph             # PAW projections
        self.psit_hG = psit_hG   # Pseudo wave function plane-wave components
        self.h_myt = h_myt       # myt -> h index mapping

    @lazyproperty
    def nh(self):
        nh = len(self.eps_h)
        assert len(self.f_h) == nh
        assert self.Ph.nbands == nh
        assert len(self.psit_hG) == nh

        return nh

    @property
    def eps_myt(self):
        return self.eps_h[self.h_myt]

    @property
    def f_myt(self):
        return self.f_h[self.h_myt]

    def projectors_in_transition_index(self, Ph):
        Pmyt = Ph.new(nbands=len(self.h_myt), bcomm=None)
        Pmyt.array[:] = Ph.array[self.h_myt]
        return Pmyt


class KohnShamKPointPair:
    """Data of pairs of Kohn-Sham orbital pertaining to transitions k -> k'."""

    def __init__(self, K1, K2, ikpt1, ikpt2, transitions, tblocks):
        """Construct the KohnShamKPointPair from the k-point data of k and k'.

        K1, K2 : int, int
            k-point indices of k and k'
        ikpt1, ikpt2 : IrreducibleKPoint, IrreducibleKPoint
            k-point data of the two specific k-points in the irreducible part
            of the BZ which are related to K1 and K2 by symmetry respectively.
        """

        self.K1 = K1
        self.K2 = K2
        self.ikpt1 = ikpt1
        self.ikpt2 = ikpt2
        self.transitions = transitions
        self.tblocks = tblocks

    def get_all(self, in_mytx):
        """Get a certain data array with all transitions"""
        return self.tblocks.all_gather(in_mytx)

    @property
    def deps_t(self):
        return self.get_all(self.ikpt2.eps_myt) \
            - self.get_all(self.ikpt1.eps_myt)

    @property
    def df_t(self):
        return self.get_all(self.ikpt2.f_myt) \
            - self.get_all(self.ikpt1.f_myt)


class KohnShamKPointPairExtractor:
    """Functionality to extract KohnShamKPointPairs from a
    ResponseGroundStateAdapter."""

    def __init__(self, gs, context, *,
                 transitions_blockcomm, kpts_blockcomm):
        """
        Parameters
        ----------
        gs : ResponseGroundStateAdapter
        context : ResponseContext
        transitions_blockcomm : gpaw.mpi.Communicator
            Communicator to distribute band and spin transitions
        kpts_blockcomm : gpaw.mpi.Communicator
            Communicator over which the k-point are distributed
        """
        assert isinstance(gs, ResponseGroundStateAdapter)
        self.gs = gs
        assert isinstance(context, ResponseContext)
        self.context = context

        self.calc_parallel = self.check_calc_parallelisation()

        self.transitions_blockcomm = transitions_blockcomm
        self.kpts_blockcomm = kpts_blockcomm

        # Prepare to distribute transitions
        self.tblocks = None

        # Prepare to find k-point data from vector
        kd = self.gs.kd
        self.kptfinder = KPointFinder(kd.bzk_kc)

        # Prepare to redistribute kptdata
        self.rrequests = []
        self.srequests = []

        # Count bands so it is possible to remove null transitions
        self.nocc1 = None  # number of completely filled bands
        self.nocc2 = None  # number of non-empty bands
        self.count_occupied_bands()

    def check_calc_parallelisation(self):
        """Check how ground state calculation is distributed in memory"""
        if self.gs.world.size == 1:
            return False
        else:
            assert self.context.comm.rank == self.gs.world.rank
            assert self.gs.gd.comm.size == 1
            return True

    def count_occupied_bands(self):
        """Count number of occupied and unoccupied bands in ground state
        calculation. Can be used to omit null-transitions between two occupied
        bands or between two unoccupied bands."""

        nocc1, nocc2 = self.gs.count_occupied_bands(ftol=1e-9)
        nocc1 = int(nocc1)
        nocc2 = int(nocc2)

        # Collect nocc for all k-points
        nocc1 = self.gs.kd.comm.min(nocc1)
        nocc2 = self.gs.kd.comm.max(nocc2)

        # Sum over band distribution
        nocc1 = self.gs.bd.comm.sum(nocc1)
        nocc2 = self.gs.bd.comm.sum(nocc2)

        self.nocc1 = int(nocc1)
        self.nocc2 = int(nocc2)
        self.context.print('Number of completely filled bands:',
                           self.nocc1, flush=False)
        self.context.print('Number of partially filled bands:',
                           self.nocc2, flush=False)
        self.context.print('Total number of bands:', self.gs.bd.nbands)

    @timer('Get Kohn-Sham pairs')
    def get_kpoint_pairs(self, k1_pc, k2_pc,
                         transitions) -> KohnShamKPointPair | None:
        """Get all pairs of Kohn-Sham orbitals for transitions k -> k'

        (n1_t, k1_p, s1_t) -> (n2_t, k2_p, s2_t)

        Here, t is a composite band and spin transition index accounted for by
        the input PairTransitions object, whereas p indexes the k-point that
        each rank of the k-point block communicator needs to extract."""
        assert k1_pc.shape == k2_pc.shape

        # Distribute transitions and extract data for transitions in
        # this process' block
        self.tblocks = Blocks1D(self.transitions_blockcomm, len(transitions))

        K1, ikpt1 = self.get_kpoints(k1_pc, transitions.n1_t, transitions.s1_t)
        K2, ikpt2 = self.get_kpoints(k2_pc, transitions.n2_t, transitions.s2_t)

        # The process might not have a Kohn-Sham k-point pair to return, due to
        # the distribution over kpts_blockcomm
        if self.kpts_blockcomm.rank not in range(len(k1_pc)):
            return None

        assert K1 is not None and ikpt1 is not None
        assert K2 is not None and ikpt2 is not None

        return KohnShamKPointPair(K1, K2, ikpt1, ikpt2,
                                  transitions, self.tblocks)

    def get_kpoints(self, k_pc, n_t, s_t):
        """Get the process' own k-point data and help other processes
        extracting theirs."""
        assert len(n_t) == len(s_t)
        assert len(k_pc) <= self.kpts_blockcomm.size

        # Use the data extraction factory to extract the kptdata
        kptdata = self.extract_kptdata(k_pc, n_t, s_t)

        if self.kpts_blockcomm.rank not in range(len(k_pc)):
            return None, None  # The process has no data of its own

        assert kptdata is not None
        K = kptdata[0]
        ikpt = IrreducibleKPoint(*kptdata[1:])

        return K, ikpt

    @timer('Extracting data from the ground state calculator object')
    def extract_kptdata(self, k_pc, n_t, s_t):
        """Extract the input data needed to construct the IrreducibleKPoints.
        """
        if self.calc_parallel:
            return self.parallel_extract_kptdata(k_pc, n_t, s_t)
        else:
            return self.serial_extract_kptdata(k_pc, n_t, s_t)
            # Useful for debugging:
            # return self.parallel_extract_kptdata(k_pc, n_t, s_t)

    def parallel_extract_kptdata(self, k_pc, n_t, s_t):
        """Extract the k-point data from a parallelized calculator."""
        (myK, myik, myu_eu,
         myn_eueh, ik_r2,
         nrh_r2, eh_eur2reh,
         rh_eur2reh, h_r1rh,
         h_myt) = self.get_parallel_extraction_protocol(k_pc, n_t, s_t)

        (eps_r1rh, f_r1rh,
         P_r1rhI, psit_r1rhG,
         eps_r2rh, f_r2rh,
         P_r2rhI, psit_r2rhG) = self.allocate_transfer_arrays(myik, nrh_r2,
                                                              ik_r2, h_r1rh)

        # Do actual extraction
        for myu, myn_eh, eh_r2reh, rh_r2reh in zip(myu_eu, myn_eueh,
                                                   eh_eur2reh, rh_eur2reh):

            eps_eh, f_eh, P_ehI = self.extract_wfs_data(myu, myn_eh)

            for r2, (eh_reh, rh_reh) in enumerate(zip(eh_r2reh, rh_r2reh)):
                if eh_reh:
                    eps_r2rh[r2][rh_reh] = eps_eh[eh_reh]
                    f_r2rh[r2][rh_reh] = f_eh[eh_reh]
                    P_r2rhI[r2][rh_reh] = P_ehI[eh_reh]

            # Wavefunctions are heavy objects which can only be extracted
            # for one band index at a time, handle them seperately
            self.add_wave_function(myu, myn_eh, eh_r2reh,
                                   rh_r2reh, psit_r2rhG)

        self.distribute_extracted_data(eps_r1rh, f_r1rh, P_r1rhI, psit_r1rhG,
                                       eps_r2rh, f_r2rh, P_r2rhI, psit_r2rhG)

        # Some processes may not have to return a k-point
        if myik is None:
            data = None
        else:
            eps_h, f_h, Ph, psit_hG = self.collect_kptdata(
                myik, h_r1rh, eps_r1rh, f_r1rh, P_r1rhI, psit_r1rhG)
            data = myK, myik, eps_h, f_h, Ph, psit_hG, h_myt

        # Wait for communication to finish
        with self.context.timer('Waiting to complete mpi.send'):
            while self.srequests:
                self.context.comm.wait(self.srequests.pop(0))

        return data

    @timer('Create data extraction protocol')
    def get_parallel_extraction_protocol(self, k_pc, n_t, s_t):
        """Figure out how to extract data efficiently in parallel."""
        comm = self.context.comm
        get_extraction_info = self.create_get_extraction_info()

        # (K, ik) for each process
        mykpt = (None, None)

        # Extraction protocol
        myu_eu = []
        myn_eueh = []

        # Data distribution protocol
        nrh_r2 = np.zeros(comm.size, dtype=int)
        ik_r2 = [None for _ in range(comm.size)]
        eh_eur2reh = []
        rh_eur2reh = []
        h_r1rh = [list([]) for _ in range(comm.size)]

        # h to t index mapping
        t_myt = self.tblocks.myslice
        n_myt, s_myt = n_t[t_myt], s_t[t_myt]
        h_myt = np.empty(self.tblocks.nlocal, dtype=int)

        nt = len(n_t)
        assert nt == len(s_t)
        t_t = np.arange(nt)
        nh = 0
        for p, k_c in enumerate(k_pc):  # p indicates the receiving process
            K = self.kptfinder.find(k_c)
            ik = self.gs.kd.bz2ibz_k[K]
            for r2 in range(p * self.tblocks.blockcomm.size,
                            min((p + 1) * self.tblocks.blockcomm.size,
                                comm.size)):
                ik_r2[r2] = ik

            if p == self.kpts_blockcomm.rank:
                mykpt = (K, ik)

            # Find out who should store the data in KSKPpoint
            r2_t, myt_t = self.map_who_has(p, t_t)

            # Find out how to extract data
            # In the ground state, kpts are indexed by u=(s, k)
            for s in set(s_t):
                thiss_myt = s_myt == s
                thiss_t = s_t == s
                t_ct = t_t[thiss_t]
                n_ct = n_t[thiss_t]
                r2_ct = r2_t[t_ct]

                # Find out where data is in GS
                u = ik * self.gs.nspins + s
                myu, r1_ct, myn_ct = get_extraction_info(u, n_ct, r2_ct)

                # If the process is extracting or receiving data,
                # figure out how to do so
                if comm.rank in np.append(r1_ct, r2_ct):
                    # Does this process have anything to send?
                    thisr1_ct = r1_ct == comm.rank
                    if np.any(thisr1_ct):
                        eh_r2reh = [list([]) for _ in range(comm.size)]
                        rh_r2reh = [list([]) for _ in range(comm.size)]
                        # Find composite indeces h = (n, s)
                        n_et = n_ct[thisr1_ct]
                        n_eh = np.unique(n_et)
                        # Find composite local band indeces
                        myn_eh = np.unique(myn_ct[thisr1_ct])

                        # Where to send the data
                        r2_et = r2_ct[thisr1_ct]
                        for r2 in np.unique(r2_et):
                            thisr2_et = r2_et == r2
                            # What ns are the process sending?
                            n_reh = np.unique(n_et[thisr2_et])
                            eh_reh = []
                            for n in n_reh:
                                eh_reh.append(np.where(n_eh == n)[0][0])
                            # How to send it
                            eh_r2reh[r2] = eh_reh
                            nreh = len(eh_reh)
                            rh_r2reh[r2] = np.arange(nreh) + nrh_r2[r2]
                            nrh_r2[r2] += nreh

                        myu_eu.append(myu)
                        myn_eueh.append(myn_eh)
                        eh_eur2reh.append(eh_r2reh)
                        rh_eur2reh.append(rh_r2reh)

                    # Does this process have anything to receive?
                    thisr2_ct = r2_ct == comm.rank
                    if np.any(thisr2_ct):
                        # Find unique composite indeces h = (n, s)
                        n_rt = n_ct[thisr2_ct]
                        n_rn = np.unique(n_rt)
                        nrn = len(n_rn)
                        h_rn = np.arange(nrn) + nh
                        nh += nrn

                        # Where to get the data from
                        r1_rt = r1_ct[thisr2_ct]
                        for r1 in np.unique(r1_rt):
                            thisr1_rt = r1_rt == r1
                            # What ns are the process getting?
                            n_reh = np.unique(n_rt[thisr1_rt])
                            # Where to put them
                            for n in n_reh:
                                h = h_rn[np.where(n_rn == n)[0][0]]
                                h_r1rh[r1].append(h)

                                # h to t mapping
                                thisn_myt = n_myt == n
                                thish_myt = np.logical_and(thisn_myt,
                                                           thiss_myt)
                                h_myt[thish_myt] = h

        return (*mykpt, myu_eu, myn_eueh, ik_r2, nrh_r2,
                eh_eur2reh, rh_eur2reh, h_r1rh, h_myt)

    def create_get_extraction_info(self):
        """Creator component of the extraction information factory."""
        if self.calc_parallel:
            return self.get_parallel_extraction_info
        else:
            return self.get_serial_extraction_info

    @staticmethod
    def get_serial_extraction_info(u, n_ct, r2_ct):
        """Figure out where to extract the data from in the gs calc"""
        # Let the process extract its own data
        myu = u  # The process has access to all data
        r1_ct = r2_ct
        myn_ct = n_ct

        return myu, r1_ct, myn_ct

    def get_parallel_extraction_info(self, u, n_ct, *unused):
        """Figure out where to extract the data from in the gs calc"""
        gs = self.gs
        # Find out where data is in GS
        k, s = divmod(u, gs.nspins)
        kptrank, q = gs.kd.who_has(k)
        myu = q * gs.nspins + s
        r1_ct, myn_ct = [], []
        for n in n_ct:
            bandrank, myn = gs.bd.who_has(n)
            # XXX this will fail when using non-standard nesting
            # of communicators.
            r1 = (kptrank * gs.gd.comm.size * gs.bd.comm.size
                  + bandrank * gs.gd.comm.size)
            r1_ct.append(r1)
            myn_ct.append(myn)

        return myu, np.array(r1_ct), np.array(myn_ct)

    @timer('Allocate transfer arrays')
    def allocate_transfer_arrays(self, myik, nrh_r2, ik_r2, h_r1rh):
        """Allocate arrays for intermediate storage of data."""
        kptex = self.gs.kpt_u[0]
        Pshape = kptex.projections.array.shape
        Pdtype = kptex.projections.matrix.dtype
        psitdtype = kptex.psit.array.dtype

        # Number of h-indeces to receive
        nrh_r1 = [len(h_rh) for h_rh in h_r1rh]

        # if self.kpts_blockcomm.rank in range(len(ik_p)):
        if myik is not None:
            ng = self.gs.global_pd.ng_q[myik]
            eps_r1rh, f_r1rh, P_r1rhI, psit_r1rhG = [], [], [], []
            for nrh in nrh_r1:
                if nrh >= 1:
                    eps_r1rh.append(np.empty(nrh))
                    f_r1rh.append(np.empty(nrh))
                    P_r1rhI.append(np.empty((nrh,) + Pshape[1:], dtype=Pdtype))
                    psit_r1rhG.append(np.empty((nrh, ng), dtype=psitdtype))
                else:
                    eps_r1rh.append(None)
                    f_r1rh.append(None)
                    P_r1rhI.append(None)
                    psit_r1rhG.append(None)
        else:
            eps_r1rh, f_r1rh, P_r1rhI, psit_r1rhG = None, None, None, None

        eps_r2rh, f_r2rh, P_r2rhI, psit_r2rhG = [], [], [], []
        for nrh, ik in zip(nrh_r2, ik_r2):
            if nrh:
                eps_r2rh.append(np.empty(nrh))
                f_r2rh.append(np.empty(nrh))
                P_r2rhI.append(np.empty((nrh,) + Pshape[1:], dtype=Pdtype))
                ng = self.gs.global_pd.ng_q[ik]
                psit_r2rhG.append(np.empty((nrh, ng), dtype=psitdtype))
            else:
                eps_r2rh.append(None)
                f_r2rh.append(None)
                P_r2rhI.append(None)
                psit_r2rhG.append(None)

        return (eps_r1rh, f_r1rh, P_r1rhI, psit_r1rhG,
                eps_r2rh, f_r2rh, P_r2rhI, psit_r2rhG)

    def map_who_has(self, p, t_t):
        """Convert k-point and transition index to global world rank
        and local transition index"""
        trank_t, myt_t = np.divmod(t_t, self.tblocks.blocksize)
        return p * self.tblocks.blockcomm.size + trank_t, myt_t

    @timer('Extracting eps, f and P_I from wfs')
    def extract_wfs_data(self, myu, myn_eh):
        kpt = self.gs.kpt_u[myu]
        # Get eig and occ
        eps_eh, f_eh = kpt.eps_n[myn_eh], kpt.f_n[myn_eh] / kpt.weight

        # Get projections
        assert kpt.projections.atom_partition.comm.size == 1
        P_ehI = kpt.projections.array[myn_eh]

        return eps_eh, f_eh, P_ehI

    @timer('Extracting wave function from wfs')
    def add_wave_function(self, myu, myn_eh,
                          eh_r2reh, rh_r2reh, psit_r2rhG):
        """Add the plane wave coefficients of the smooth part of
        the wave function to the psit_r2rtG arrays."""
        kpt = self.gs.kpt_u[myu]

        for eh_reh, rh_reh, psit_rhG in zip(eh_r2reh, rh_r2reh, psit_r2rhG):
            if eh_reh:
                for eh, rh in zip(eh_reh, rh_reh):
                    psit_rhG[rh] = kpt.psit_nG[myn_eh[eh]]

    @timer('Distributing kptdata')
    def distribute_extracted_data(self, eps_r1rh, f_r1rh, P_r1rhI, psit_r1rhG,
                                  eps_r2rh, f_r2rh, P_r2rhI, psit_r2rhG):
        """Send the extracted data to appropriate destinations"""
        comm = self.context.comm
        # Store the data extracted by the process itself
        rank = comm.rank
        # Check if there is actually some data to store
        if eps_r2rh[rank] is not None:
            eps_r1rh[rank] = eps_r2rh[rank]
            f_r1rh[rank] = f_r2rh[rank]
            P_r1rhI[rank] = P_r2rhI[rank]
            psit_r1rhG[rank] = psit_r2rhG[rank]

        # Receive data
        if eps_r1rh is not None:  # The process may not be receiving anything
            for r1, (eps_rh, f_rh,
                     P_rhI, psit_rhG) in enumerate(zip(eps_r1rh, f_r1rh,
                                                       P_r1rhI, psit_r1rhG)):
                # Check if there is any data to receive
                if r1 != rank and eps_rh is not None:
                    rreq1 = comm.receive(eps_rh, r1, tag=201, block=False)
                    rreq2 = comm.receive(f_rh, r1, tag=202, block=False)
                    rreq3 = comm.receive(P_rhI, r1, tag=203, block=False)
                    rreq4 = comm.receive(psit_rhG, r1, tag=204, block=False)
                    self.rrequests += [rreq1, rreq2, rreq3, rreq4]

        # Send data
        for r2, (eps_rh, f_rh,
                 P_rhI, psit_rhG) in enumerate(zip(eps_r2rh, f_r2rh,
                                                   P_r2rhI, psit_r2rhG)):
            # Check if there is any data to send
            if r2 != rank and eps_rh is not None:
                sreq1 = comm.send(eps_rh, r2, tag=201, block=False)
                sreq2 = comm.send(f_rh, r2, tag=202, block=False)
                sreq3 = comm.send(P_rhI, r2, tag=203, block=False)
                sreq4 = comm.send(psit_rhG, r2, tag=204, block=False)
                self.srequests += [sreq1, sreq2, sreq3, sreq4]

        with self.context.timer('Waiting to complete mpi.receive'):
            while self.rrequests:
                comm.wait(self.rrequests.pop(0))

    @timer('Collecting kptdata')
    def collect_kptdata(self, myik, h_r1rh,
                        eps_r1rh, f_r1rh, P_r1rhI, psit_r1rhG):
        """From the extracted data, collect the IrreducibleKPoint data arrays
        """
        kpt0 = self.gs.kpt_u[0]
        # Allocate data arrays
        maxh_r1 = [max(h_rh) for h_rh in h_r1rh if h_rh]
        if maxh_r1:
            nh = max(maxh_r1) + 1
            Ph = kpt0.projections.new(nbands=nh, bcomm=None)
        else:  # Carry around empty array
            assert self.tblocks.a == self.tblocks.b
            nh = 0
            # We have to initialize the projections by hand, because
            # Projections.new() interprets nbands == 0 to imply that it should
            # inherit the preexisting number of bands...
            proj = kpt0.projections
            Ph = Projections(nh, proj.nproj_a, proj.atom_partition,
                             serial_comm, proj.collinear, proj.spin,
                             proj.matrix.dtype)
        eps_h = np.empty(nh)
        f_h = np.empty(nh)
        assert self.gs.dtype == kpt0.psit.array.dtype
        psit_hG = np.empty((nh, self.gs.global_pd.ng_q[myik]), self.gs.dtype)

        # Store extracted data in the arrays
        for (h_rh, eps_rh,
             f_rh, P_rhI, psit_rhG) in zip(h_r1rh, eps_r1rh,
                                           f_r1rh, P_r1rhI, psit_r1rhG):
            if h_rh:
                eps_h[h_rh] = eps_rh
                f_h[h_rh] = f_rh
                Ph.array[h_rh] = P_rhI
                psit_hG[h_rh] = psit_rhG

        return eps_h, f_h, Ph, psit_hG

    def serial_extract_kptdata(self, k_pc, n_t, s_t):
        """Extract the k-point data from a serial calculator.

        Since all the processes can access all of the data, each process
        extracts the data of its own k-point without any need for
        communication."""
        if self.kpts_blockcomm.rank not in range(len(k_pc)):
            # No data to extract
            return None

        gs = self.gs
        kpt_u = gs.kpt_u

        # Find k-point indeces
        k_c = k_pc[self.kpts_blockcomm.rank]
        K = self.kptfinder.find(k_c)
        ik = gs.kd.bz2ibz_k[K]

        (myu_eu, myn_eurn, nh,
         h_eurn, h_myt) = self.get_serial_extraction_protocol(ik, n_t, s_t)

        # Allocate transfer arrays
        eps_h = np.empty(nh)
        f_h = np.empty(nh)
        Ph = kpt_u[0].projections.new(nbands=nh, bcomm=None)
        psit_hG = np.empty((nh, gs.pd.ng_q[ik]),
                           dtype=kpt_u[0].psit.array.dtype)

        # Extract data from the ground state
        for myu, myn_rn, h_rn in zip(myu_eu, myn_eurn, h_eurn):
            kpt = kpt_u[myu]
            with self.context.timer('Extracting eps, f and P_I from wfs'):
                eps_h[h_rn] = kpt.eps_n[myn_rn]
                f_h[h_rn] = kpt.f_n[myn_rn] / kpt.weight
                Ph.array[h_rn] = kpt.projections.array[myn_rn]

            with self.context.timer('Extracting wave function from wfs'):
                for myn, h in zip(myn_rn, h_rn):
                    psit_hG[h] = kpt.psit_nG[myn]

        return K, ik, eps_h, f_h, Ph, psit_hG, h_myt

    @timer('Create data extraction protocol')
    def get_serial_extraction_protocol(self, ik, n_t, s_t):
        """Figure out how to extract data efficiently in serial."""

        # Only extract the transitions handled by the process itself
        t_myt = self.tblocks.myslice
        n_myt = n_t[t_myt]
        s_myt = s_t[t_myt]

        # In the ground state, kpts are indexed by u=(s, k)
        myu_eu = []
        myn_eurn = []
        nh = 0
        h_eurn = []
        h_myt = np.empty(self.tblocks.nlocal, dtype=int)
        for s in set(s_myt):
            thiss_myt = s_myt == s
            n_ct = n_myt[thiss_myt]

            # Find unique composite h = (n, u) indeces
            n_rn = np.unique(n_ct)
            nrn = len(n_rn)
            h_eurn.append(np.arange(nrn) + nh)
            nh += nrn

            # Find mapping between h and the transition index
            for n, h in zip(n_rn, h_eurn[-1]):
                thisn_myt = n_myt == n
                thish_myt = np.logical_and(thisn_myt, thiss_myt)
                h_myt[thish_myt] = h

            # Find out where data is
            u = ik * self.gs.nspins + s
            # The process has access to all data
            myu = u
            myn_rn = n_rn

            myu_eu.append(myu)
            myn_eurn.append(myn_rn)

        return myu_eu, myn_eurn, nh, h_eurn, h_myt
